# experiment.py: Main class for an experiment (nkululeko.nkululeko)
import ast
import os
import pickle
import random
import time

import audeer
import audformat
import numpy as np
import pandas as pd
from sklearn.preprocessing import LabelEncoder

import nkululeko.glob_conf as glob_conf
from nkululeko.data.dataset import Dataset
from nkululeko.data.dataset_csv import Dataset_CSV
from nkululeko.demo_predictor import Demo_predictor
from nkululeko.feat_extract.feats_analyser import FeatureAnalyser
from nkululeko.feature_extractor import FeatureExtractor
from nkululeko.file_checker import FileChecker
from nkululeko.filter_data import DataFilter
from nkululeko.plots import Plots
from nkululeko.reporting.report import Report
from nkululeko.runmanager import Runmanager
from nkululeko.scaler import Scaler
from nkululeko.test_predictor import TestPredictor
from nkululeko.utils.util import Util


class Experiment:
    """Main class specifying an experiment."""

    def __init__(self, config_obj):
        """Constructor.

        Args:
            - config_obj : a config parser object that sets the experiment parameters and being set as a global object.
        """
        self.set_globals(config_obj)
        self.name = glob_conf.config["EXP"]["name"]
        self.root = os.path.join(glob_conf.config["EXP"]["root"], "")
        self.data_dir = os.path.join(self.root, self.name)
        audeer.mkdir(self.data_dir)  # create the experiment directory
        self.util = Util("experiment")
        glob_conf.set_util(self.util)
        fresh_report = eval(self.util.config_val("REPORT", "fresh", "False"))
        if not fresh_report:
            try:
                with open(os.path.join(self.data_dir, "report.pkl"), "rb") as handle:
                    self.report = pickle.load(handle)
            except FileNotFoundError:
                self.report = Report()
        else:
            self.util.debug("starting a fresh report")
            self.report = Report()
        glob_conf.set_report(self.report)
        self.loso = self.util.config_val("MODEL", "loso", False)
        self.logo = self.util.config_val("MODEL", "logo", False)
        self.xfoldx = self.util.config_val("MODEL", "k_fold_cross", False)
        self.start = time.process_time()

    def set_module(self, module):
        glob_conf.set_module(module)

    def store_report(self):
        with open(os.path.join(self.data_dir, "report.pkl"), "wb") as handle:
            pickle.dump(self.report, handle)
        if eval(self.util.config_val("REPORT", "show", "False")):
            self.report.print()
        if self.util.config_val("REPORT", "latex", False):
            self.report.export_latex()

    # moved to util
    # def get_name(self):
    #     return self.util.get_exp_name()

    def set_globals(self, config_obj):
        """install a config object in the global space"""
        glob_conf.init_config(config_obj)

    def load_datasets(self):
        """Load all databases specified in the configuration and map the labels"""
        ds = ast.literal_eval(glob_conf.config["DATA"]["databases"])
        self.datasets = {}
        self.got_speaker, self.got_gender, self.got_age = False, False, False
        for d in ds:
            ds_type = self.util.config_val_data(d, "type", "audformat")
            if ds_type == "audformat":
                data = Dataset(d)
            elif ds_type == "csv":
                data = Dataset_CSV(d)
            else:
                self.util.error(f"unknown data type: {ds_type}")
            data.load()
            data.prepare()
            if data.got_gender:
                self.got_gender = True
            if data.got_age:
                self.got_age = True
            if data.got_speaker:
                self.got_speaker = True
            self.datasets.update({d: data})
        self.target = self.util.config_val("DATA", "target", "none")
        glob_conf.set_target(self.target)
        # print target via debug
        self.util.debug(f"target: {self.target}")
        # print keys/column
        dbs = ",".join(list(self.datasets.keys()))
        if self.target == "none":
            self.util.debug(f"loaded databases {dbs}")
            return
        labels = self.util.config_val("DATA", "labels", False)
        auto_labels = list(next(iter(self.datasets.values())).df[self.target].unique())
        if labels:
            self.labels = ast.literal_eval(labels)
            self.util.debug(f"Using target labels (from config): {labels}")
        else:
            self.labels = auto_labels
        # print autolabel no matter it is specified or not
        self.util.debug(f"Labels (from database): {auto_labels}")
        glob_conf.set_labels(self.labels)
        self.util.debug(f"loaded databases {dbs}")

    def _import_csv(self, storage):
        # df = pd.read_csv(storage, header=0, index_col=[0,1,2])
        # df.index.set_levels(pd.to_timedelta(df.index.levels[1]), level=1)
        # df.index.set_levels(pd.to_timedelta(df.index.levels[2]), level=2)
        df = audformat.utils.read_csv(storage)
        df.is_labeled = True if self.target in df else False
        # print(df.head())
        return df

    def fill_tests(self):
        """Only fill a new test set"""

        test_dbs = ast.literal_eval(glob_conf.config["DATA"]["tests"])
        self.df_test = pd.DataFrame()
        start_fresh = eval(self.util.config_val("DATA", "no_reuse", "False"))
        store = self.util.get_path("store")
        storage_test = f"{store}extra_testdf.csv"
        if os.path.isfile(storage_test) and not start_fresh:
            self.util.debug(f"reusing previously stored {storage_test}")
            self.df_test = self._import_csv(storage_test)
        else:
            for d in test_dbs:
                ds_type = self.util.config_val_data(d, "type", "audformat")
                if ds_type == "audformat":
                    data = Dataset(d)
                elif ds_type == "csv":
                    data = Dataset_CSV(d)
                else:
                    self.util.error(f"unknown data type: {ds_type}")
                data.load()
                if data.got_gender:
                    self.got_gender = True
                if data.got_age:
                    self.got_age = True
                if data.got_speaker:
                    self.got_speaker = True
                data.split()
                data.prepare_labels()
                self.df_test = pd.concat(
                    [self.df_test, self.util.make_segmented_index(data.df_test)]
                )
                self.df_test.is_labeled = data.is_labeled
            self.df_test.got_gender = self.got_gender
            self.df_test.got_speaker = self.got_speaker
            # self.util.set_config_val('FEATS', 'needs_features_extraction', 'True')
            # self.util.set_config_val('FEATS', 'no_reuse', 'True')
            self.df_test["class_label"] = self.df_test[self.target]
            self.df_test[self.target] = self.label_encoder.transform(
                self.df_test[self.target]
            )
            self.df_test.to_csv(storage_test)

    def fill_train_and_tests(self):
        """Set up train and development sets. The method should be specified in the config."""
        store = self.util.get_path("store")
        storage_test = f"{store}testdf.csv"
        storage_train = f"{store}traindf.csv"
        start_fresh = eval(self.util.config_val("DATA", "no_reuse", "False"))
        if (
            os.path.isfile(storage_train)
            and os.path.isfile(storage_test)
            and not start_fresh
        ):
            self.util.debug(
                f"reusing previously stored {storage_test} and {storage_train}"
            )
            self.df_test = self._import_csv(storage_test)
            self.df_train = self._import_csv(storage_train)
        else:
            self.df_train, self.df_test = pd.DataFrame(), pd.DataFrame()
            for d in self.datasets.values():
                d.split()
                if self.target != "none":
                    d.prepare_labels()
                if d.df_train.shape[0] == 0:
                    self.util.debug(f"warn: {d.name} train empty")
                self.df_train = pd.concat([self.df_train, d.df_train])
                # print(f"df_train: {self.df_train}")
                self.util.copy_flags(d, self.df_train)
                if d.df_test.shape[0] == 0:
                    self.util.debug(f"warn: {d.name} test empty")
                self.df_test = pd.concat([self.df_test, d.df_test])
                self.util.copy_flags(d, self.df_test)
            store = self.util.get_path("store")
            storage_test = f"{store}testdf.csv"
            storage_train = f"{store}traindf.csv"
            self.df_test.to_csv(storage_test)
            self.df_train.to_csv(storage_train)

        if self.target == "none":
            return
        self.util.copy_flags(self, self.df_test)
        self.util.copy_flags(self, self.df_train)
        # Try data checks
        datachecker = FileChecker(self.df_train)
        self.df_train = datachecker.all_checks()
        datachecker.set_data(self.df_test)
        self.df_test = datachecker.all_checks()

        # Check for filters
        filter_sample_selection = self.util.config_val(
            "DATA", "filter.sample_selection", "all"
        )
        if filter_sample_selection == "all":
            datafilter = DataFilter(self.df_train)
            self.df_train = datafilter.all_filters()
            datafilter = DataFilter(self.df_test)
            self.df_test = datafilter.all_filters()
        elif filter_sample_selection == "train":
            datafilter = DataFilter(self.df_train)
            self.df_train = datafilter.all_filters()
        elif filter_sample_selection == "test":
            datafilter = DataFilter(self.df_test)
            self.df_test = datafilter.all_filters()
        else:
            self.util.error(
                "unkown filter sample selection specifier"
                f" {filter_sample_selection}, should be [all | train | test]"
            )

        # encode the labels
        if self.util.exp_is_classification():
            datatype = self.util.config_val("DATA", "type", "dummy")
            if datatype == "continuous":
                # if self.df_test.is_labeled:
                #     # remember the target in case they get labelencoded later
                #     self.df_test["class_label"] = self.df_test[self.target]
                test_cats = self.df_test["class_label"].unique()
                # else:
                #     # if there is no target, copy a dummy label
                #     self.df_test = self._add_random_target(self.df_test)
                # if self.df_train.is_labeled:
                #     # remember the target in case they get labelencoded later
                #     self.df_train["class_label"] = self.df_train[self.target]
                train_cats = self.df_train["class_label"].unique()

            else:
                if self.df_test.is_labeled:
                    test_cats = self.df_test[self.target].unique()
                else:
                    # if there is no target, copy a dummy label
                    self.df_test = self._add_random_target(self.df_test).astype("str")
                train_cats = self.df_train[self.target].unique()
                # print(f"df_train: {pd.DataFrame(self.df_train[self.target])}")
                # print(f"train_cats with target {self.target}: {train_cats}")
            if self.df_test.is_labeled:
                if type(test_cats) == np.ndarray:
                    self.util.debug(f"Categories test (nd.array): {test_cats}")
                else:
                    self.util.debug(f"Categories test (list): {list(test_cats)}")
            if type(train_cats) == np.ndarray:
                self.util.debug(f"Categories train (nd.array): {train_cats}")
            else:
                self.util.debug(f"Categories train (list): {list(train_cats)}")

            # encode the labels as numbers
            self.label_encoder = LabelEncoder()
            self.df_train[self.target] = self.label_encoder.fit_transform(
                self.df_train[self.target]
            )
            self.df_test[self.target] = self.label_encoder.transform(
                self.df_test[self.target]
            )
            glob_conf.set_label_encoder(self.label_encoder)
        if self.got_speaker:
            self.util.debug(
                f"{self.df_test.speaker.nunique()} speakers in test and"
                f" {self.df_train.speaker.nunique()} speakers in train"
            )

        target_factor = self.util.config_val("DATA", "target_divide_by", False)
        if target_factor:
            self.df_test[self.target] = self.df_test[self.target] / float(target_factor)
            self.df_train[self.target] = self.df_train[self.target] / float(
                target_factor
            )
            if not self.util.exp_is_classification():
                self.df_test["class_label"] = self.df_test["class_label"] / float(
                    target_factor
                )
                self.df_train["class_label"] = self.df_train["class_label"] / float(
                    target_factor
                )

    def _add_random_target(self, df):
        labels = glob_conf.labels
        a = [None] * len(df)
        for i in range(0, len(df)):
            a[i] = random.choice(labels)
        df[self.target] = a
        return df

    def plot_distribution(self, df_labels):
        """Plot the distribution of samples and speaker per target class and biological sex"""
        plot = Plots()
        sample_selection = self.util.config_val("EXPL", "sample_selection", "all")
        plot.plot_distributions(df_labels)
        if self.got_speaker:
            plot.plot_distributions_speaker(df_labels)

    def extract_test_feats(self):
        self.feats_test = pd.DataFrame()
        feats_name = "_".join(ast.literal_eval(glob_conf.config["DATA"]["tests"]))
        feats_types = self.util.config_val_list("FEATS", "type", ["os"])
        self.feature_extractor = FeatureExtractor(
            self.df_test, feats_types, feats_name, "test"
        )
        self.feats_test = self.feature_extractor.extract()
        self.util.debug(f"Test features shape:{self.feats_test.shape}")

    def extract_feats(self):
        """Extract the features for train and dev sets.

        They will be stored on disk and need to be removed manually.

        The string FEATS.feats_type is read from the config, defaults to os.

        """
        df_train, df_test = self.df_train, self.df_test
        feats_name = "_".join(ast.literal_eval(glob_conf.config["DATA"]["databases"]))
        self.feats_test, self.feats_train = pd.DataFrame(), pd.DataFrame()
        feats_types = self.util.config_val("FEATS", "type", "os")
        # Ensure feats_types is always a list of strings
        if isinstance(feats_types, str):
            if feats_types.startswith("[") and feats_types.endswith("]"):
                feats_types = ast.literal_eval(feats_types)
            else:
                feats_types = [feats_types]
        # print(f"feats_types: {feats_types}")
        # for some models no features are needed
        if len(feats_types) == 0:
            self.util.debug("no feature extractor specified.")
            self.feats_train, self.feats_test = pd.DataFrame(), pd.DataFrame()
            return
        self.feature_extractor = FeatureExtractor(
            df_train, feats_types, feats_name, "train"
        )
        self.feats_train = self.feature_extractor.extract()
        self.feature_extractor = FeatureExtractor(
            df_test, feats_types, feats_name, "test"
        )
        self.feats_test = self.feature_extractor.extract()
        self.util.debug(
            f"All features: train shape : {self.feats_train.shape}, test"
            f" shape:{self.feats_test.shape}"
        )
        if self.feats_train.shape[0] < self.df_train.shape[0]:
            self.util.warn(
                f"train feats ({self.feats_train.shape[0]}) != train labels"
                f" ({self.df_train.shape[0]})"
            )
            self.df_train = self.df_train[
                self.df_train.index.isin(self.feats_train.index)
            ]
            self.util.warn(f"new train labels shape: {self.df_train.shape[0]}")
        if self.feats_test.shape[0] < self.df_test.shape[0]:
            self.util.warn(
                f"test feats ({self.feats_test.shape[0]}) != test labels"
                f" ({self.df_test.shape[0]})"
            )
            self.df_test = self.df_test[self.df_test.index.isin(self.feats_test.index)]
            self.util.warn(f"new test labels shape: {self.df_test.shape[0]}")

        self._check_scale()
        # store = self.util.get_path("store")
        # store_format = self.util.config_val("FEATS", "store_format", "pkl")
        # storage = f"{store}test_feats.{store_format}"
        # self.util.write_store(self.feats_test, storage, store_format)
        # storage = f"{store}train_feats.{store_format}"
        # self.util.write_store(self.feats_train, storage, store_format)

    def augment(self):
        """Augment the selected samples."""
        from nkululeko.augmenting.augmenter import Augmenter

        sample_selection = self.util.config_val("AUGMENT", "sample_selection", "all")
        if sample_selection == "all":
            df = pd.concat([self.df_train, self.df_test])
        elif sample_selection == "train":
            df = self.df_train
        elif sample_selection == "test":
            df = self.df_test
        else:
            self.util.error(
                f"unknown augmentation selection specifier {sample_selection},"
                " should be [all | train | test]"
            )

        augmenter = Augmenter(df)
        df_ret = augmenter.augment(sample_selection)
        return df_ret

    def autopredict(self):
        """
        Predict labels for samples with existing models and add to the dataframe.
        """
        sample_selection = self.util.config_val("PREDICT", "split", "all")
        if sample_selection == "all":
            df = pd.concat([self.df_train, self.df_test])
        elif sample_selection == "train":
            df = self.df_train
        elif sample_selection == "test":
            df = self.df_test
        else:
            self.util.error(
                f"unknown augmentation selection specifier {sample_selection},"
                " should be [all | train | test]"
            )
        targets = self.util.config_val_list("PREDICT", "targets", ["gender"])
        for target in targets:
            if target == "gender":
                from nkululeko.autopredict.ap_gender import GenderPredictor

                predictor = GenderPredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "age":
                from nkululeko.autopredict.ap_age import AgePredictor

                predictor = AgePredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "snr":
                from nkululeko.autopredict.ap_snr import SNRPredictor

                predictor = SNRPredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "mos":
                from nkululeko.autopredict.ap_mos import MOSPredictor

                predictor = MOSPredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "pesq":
                from nkululeko.autopredict.ap_pesq import PESQPredictor

                predictor = PESQPredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "sdr":
                from nkululeko.autopredict.ap_sdr import SDRPredictor

                predictor = SDRPredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "stoi":
                from nkululeko.autopredict.ap_stoi import STOIPredictor

                predictor = STOIPredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "arousal":
                from nkululeko.autopredict.ap_arousal import ArousalPredictor

                predictor = ArousalPredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "valence":
                from nkululeko.autopredict.ap_valence import ValencePredictor

                predictor = ValencePredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "dominance":
                from nkululeko.autopredict.ap_dominance import DominancePredictor

                predictor = DominancePredictor(df)
                df = predictor.predict(sample_selection)
            else:
                self.util.error(f"unknown auto predict target: {target}")
        return df

    def random_splice(self):
        """
        Random-splice the selected samples
        """
        from nkululeko.augmenting.randomsplicer import Randomsplicer

        sample_selection = self.util.config_val("AUGMENT", "sample_selection", "all")
        if sample_selection == "all":
            df = pd.concat([self.df_train, self.df_test])
        elif sample_selection == "train":
            df = self.df_train
        elif sample_selection == "test":
            df = self.df_test
        else:
            self.util.error(
                f"unknown augmentation selection specifier {sample_selection},"
                " should be [all | train | test]"
            )
        randomsplicer = Randomsplicer(df)
        df_ret = randomsplicer.run(sample_selection)
        return df_ret

    def analyse_features(self, needs_feats):
        """Do a feature exploration."""
        plot_feats = eval(
            self.util.config_val("EXPL", "feature_distributions", "False")
        )
        sample_selection = self.util.config_val("EXPL", "sample_selection", "all")
        # get the data labels
        if sample_selection == "all":
            df_labels = pd.concat([self.df_train, self.df_test])
            self.util.copy_flags(self.df_train, df_labels)
        elif sample_selection == "train":
            df_labels = self.df_train
            self.util.copy_flags(self.df_train, df_labels)
        elif sample_selection == "test":
            df_labels = self.df_test
            self.util.copy_flags(self.df_test, df_labels)
        else:
            self.util.error(
                f"unknown sample selection specifier {sample_selection}, should"
                " be [all | train | test]"
            )
        self.util.debug(f"sampling selection: {sample_selection}")
        if self.util.config_val("EXPL", "value_counts", False):
            self.plot_distribution(df_labels)

        # check if data should be shown with the spotlight data visualizer
        spotlight = eval(self.util.config_val("EXPL", "spotlight", "False"))
        if spotlight:
            self.util.debug("opening spotlight tab in web browser")
            from renumics import spotlight

            spotlight.show(df_labels.reset_index())

        if not needs_feats:
            return
        # get the feature values
        if sample_selection == "all":
            df_feats = pd.concat([self.feats_train, self.feats_test])
        elif sample_selection == "train":
            df_feats = self.feats_train
        elif sample_selection == "test":
            df_feats = self.feats_test
        else:
            self.util.error(
                f"unknown sample selection specifier {sample_selection}, should"
                " be [all | train | test]"
            )
        feat_analyser = FeatureAnalyser(sample_selection, df_labels, df_feats)
        # check if SHAP features should be analysed
        shap = eval(self.util.config_val("EXPL", "shap", "False"))
        if shap:
            feat_analyser.analyse_shap(self.runmgr.get_best_model())

        if plot_feats:
            feat_analyser.analyse()

        # check if a scatterplot should be done
        scatter_var = eval(self.util.config_val("EXPL", "scatter", "False"))
        scatter_target = self.util.config_val(
            "EXPL", "scatter.target", "['class_label']"
        )
        if scatter_var:
            scatters = ast.literal_eval(glob_conf.config["EXPL"]["scatter"])
            scat_targets = ast.literal_eval(scatter_target)
            plots = Plots()
            for scat_target in scat_targets:
                if self.util.is_categorical(df_labels[scat_target]):
                    for scatter in scatters:
                        plots.scatter_plot(df_feats, df_labels, scat_target, scatter)
                else:
                    self.util.debug(
                        f"{self.name}: binning continuous variable to categories"
                    )
                    cat_vals = self.util.continuous_to_categorical(
                        df_labels[scat_target]
                    )
                    df_labels[f"{scat_target}_bins"] = cat_vals.values
                    for scatter in scatters:
                        plots.scatter_plot(
                            df_feats, df_labels, f"{scat_target}_bins", scatter
                        )

    def _check_scale(self):
        self.util.save_to_store(self.feats_train, "feats_train")
        self.util.save_to_store(self.feats_test, "feats_test")
        scale_feats = self.util.config_val("FEATS", "scale", False)
        # print the scale
        self.util.debug(f"scaler: {scale_feats}")
        if scale_feats:
            self.scaler_feats = Scaler(
                self.df_train,
                self.df_test,
                self.feats_train,
                self.feats_test,
                scale_feats,
            )
            self.feats_train, self.feats_test = self.scaler_feats.scale()
            # store versions
            self.util.save_to_store(self.feats_train, "feats_train_scaled")
            self.util.save_to_store(self.feats_test, "feats_test_scaled")

    def init_runmanager(self):
        """Initialize the manager object for the runs."""
        self.runmgr = Runmanager(
            self.df_train, self.df_test, self.feats_train, self.feats_test
        )

    def run(self):
        """Do the runs."""
        self.runmgr.do_runs()

        # access the best results all runs
        self.reports = self.runmgr.best_results
        last_epochs = self.runmgr.last_epochs
        # try to save yourself
        save = self.util.config_val("EXP", "save", False)
        if save:
            # save the experiment for future use
            self.save(self.util.get_save_name())
            # self.save_onnx(self.util.get_save_name())

        # self.__collect_reports()
        self.util.print_best_results(self.reports)

        # check if the test predictions should be saved to disk
        test_pred_file = self.util.config_val("EXP", "save_test", False)
        if test_pred_file:
            self.predict_test_and_save(test_pred_file)

        # check if the majority voting for all speakers should be plotted
        conf_mat_per_speaker_function = self.util.config_val(
            "PLOT", "combine_per_speaker", False
        )
        if conf_mat_per_speaker_function:
            self.plot_confmat_per_speaker(conf_mat_per_speaker_function)
        used_time = time.process_time() - self.start
        self.util.debug(f"Done, used {used_time:.3f} seconds")

        # check if a test set should be labeled by the model:
        label_data = self.util.config_val("DATA", "label_data", False)
        label_result = self.util.config_val("DATA", "label_result", False)
        if label_data and label_result:
            self.predict_test_and_save(label_result)

        return self.reports, last_epochs

    def plot_confmat_per_speaker(self, function):
        if self.loso or self.logo or self.xfoldx:
            self.util.debug(
                "plot combined speaker predictions not possible for cross" " validation"
            )
            return
        best = self.get_best_report(self.reports)
        if best.is_classification:
            truths = best.truths
            preds = best.preds
        else:
            truths = best.truths_cont
            preds = best.preds_cont
        speakers = self.df_test.speaker.values
        df = pd.DataFrame(data={"truths": truths, "preds": preds, "speakers": speakers})
        plot_name = f"{self.util.get_exp_name()}_speakercombined_{function}"
        self.util.debug(
            f"plotting speaker combination ({function}) confusion matrix to"
            f" {plot_name}"
        )
        best.plot_per_speaker(df, plot_name, function)

    def get_best_report(self, reports):
        return self.runmgr.get_best_result(reports)

    def print_best_model(self):
        self.runmgr.print_best_result_runs()

    def demo(self, file, is_list, outfile):
        model = self.runmgr.get_best_model()
        lab_enc = None
        try:
            lab_enc = self.label_encoder
        except AttributeError:
            pass
        demo = Demo_predictor(
            model, file, is_list, self.feature_extractor, lab_enc, outfile
        )
        demo.run_demo()

    def predict_test_and_save(self, result_name):
        model = self.runmgr.get_best_model()
        model.set_testdata(self.df_test, self.feats_test)
        test_predictor = TestPredictor(
            model, self.df_test, self.label_encoder, result_name
        )
        result = test_predictor.predict_and_store()
        return result

    def load(self, filename):
        try:
            f = open(filename, "rb")
            tmp_dict = pickle.load(f)
            f.close()
        except EOFError as eof:
            self.util.error(f"can't open file {filename}: {eof}")
        self.__dict__.update(tmp_dict)
        glob_conf.set_labels(self.labels)

    def save(self, filename):
        if self.runmgr.modelrunner.model.is_ann():
            self.runmgr.modelrunner.model = None
            self.util.warn(
                "Save experiment: Can't pickle the trained model so saving without it. (it should be stored anyway)"
            )
        try:
            f = open(filename, "wb")
            pickle.dump(self.__dict__, f)
            f.close()
        except (TypeError, AttributeError) as error:
            self.feature_extractor.feat_extractor.model = None
            f = open(filename, "wb")
            pickle.dump(self.__dict__, f)
            f.close()
            self.util.warn(
                "Save experiment: Can't pickle the feature extraction model so saving without it."
                + f"{type(error).__name__} {error}"
            )
        except RuntimeError as error:
            self.util.warn(
                "Save experiment: Can't pickle local object, NOT saving: "
                + f"{type(error).__name__} {error}"
            )

    def save_onnx(self, filename):
        # export the model to onnx
        model = self.runmgr.get_best_model()
        if model.is_ann():
            print("converting to onnx from torch")
        else:
            print("converting to onnx from sklearn")
        # save the rest
        f = open(filename, "wb")
        pickle.dump(self.__dict__, f)
        f.close()



class DefineBase:
    @classmethod
    def _assert_has_attribute_value(cls, value):
        valid_values = cls._attribute_values()
        if value not in valid_values:
            raise BadValueError(value, valid_values)

    @classmethod
    def _attribute_values(cls):
        attributes = inspect.getmembers(cls, lambda x: not inspect.isroutine(x))
        return sorted(
            [
                a[1]
                for a in attributes
                if not (a[0].startswith("__") and a[0].endswith("__"))
            ]
        )


class Header(DefineBase):
    HEADER_RESULTS = "Results"
    HEADER_EXPLORE = "Data exploration"



import argparse
import configparser
import os

from sklearn import pipeline
from transformers import pipelines

from nkululeko.utils.util import get_exp_dir

parser = argparse.ArgumentParser()
parser.add_argument("--file", help="A file that should be processed (16kHz mono wav)")
# read config from ini file
parser.add_argument("--config", default="exp.ini", help="The base configuration")


args = parser.parse_args()
file = args.file
config = configparser.ConfigParser()

# get exp dir from config [EXP][root][name] + models + run_0 + torch
config.read(args.config)
exp_dir = get_exp_dir("model_path")

# exp_dir = get_exp_dir("model_path")
model_path = os.path.join(exp_dir, "model")
pipe = pipelines("audio-classification", model=model_path)


print(pipeline(file))



"""
report.py

Collector class for report items collected during module processing.

"""

from nkululeko.reporting.latex_writer import LatexWriter
from nkululeko.utils.util import Util


class Report:
    def __init__(self):
        self.report_items = {}
        self.util = Util("Report")
        self.initial = True

    def add_item(self, ri):
        if ri.topic in self.report_items:
            self.report_items[ri.topic].append(ri)
        else:
            self.report_items[ri.topic] = []
            self.report_items[ri.topic].append(ri)

    def print(self):
        print("###### Nkululeko Report ######")
        for topic in self.report_items:
            print(f"### {topic}")
            for c in self.report_items[topic]:
                print(c.caption)
                print("\t" + c.contents)

    def export_latex(self):
        lw = LatexWriter()
        for topic in self.report_items:
            lw.add_items_for_section(topic, self.report_items[topic])
        lw.finish_doc()



# model_gmm.py

import pandas as pd
from sklearn import mixture

from nkululeko.models.model import Model


class GMM_model(Model):
    """An GMM model"""

    is_classifier = True

    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        self.name = "gmm"
        self.n_components = int(self.util.config_val("MODEL", "GMM_components", "4"))
        covariance_type = self.util.config_val("MODEL", "GMM_covariance_type", "full")
        self.clf = mixture.GaussianMixture(
            n_components=self.n_components,
            covariance_type=covariance_type,
            random_state=42,
        )
        # set up the classifier

    def get_predictions(self):
        """Use the predict_proba method of the GaussianMixture model to get
        probabilities. Create a DataFrame with these probabilities and return
        it along with the predictions."""
        probs = self.clf.predict_proba(self.feats_test)
        preds = self.clf.predict(self.feats_test)

        # Convert predictions to a list
        preds = preds.tolist()

        # Create a DataFrame for probabilities
        proba_df = pd.DataFrame(
            probs, index=self.feats_test.index, columns=range(self.n_components)
        )

        return preds, proba_df



# export.py
# export the loaded training and test sets to it's own folder

import argparse
import configparser
import os
import shutil

import audeer
import audiofile
import pandas as pd

from nkululeko.constants import VERSION
from nkululeko.experiment import Experiment
from nkululeko.utils.util import Util


def main():
    parser = argparse.ArgumentParser(description="Call the nkululeko framework.")
    parser.add_argument("--config", default="exp.ini", help="The base configuration")
    args = parser.parse_args()
    config_file = args.config if args.config is not None else "exp.ini"

    if not os.path.isfile(config_file):
        print(f"ERROR: no such file: {config_file}")
        exit()

    config = configparser.ConfigParser()
    config.read(config_file)
    expr = Experiment(config)
    util = Util("export")
    util.debug(
        f"running {expr.name} from config {config_file}, nkululeko version"
        f" {VERSION}"
    )

    if util.config_val("EXP", "no_warnings", False):
        import warnings

        warnings.filterwarnings("ignore")

    # load the data
    expr.load_datasets()

    # split into train and test
    expr.fill_train_and_tests()
    util.debug(f"train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}")

    # export
    df_train = expr.df_train
    df_test = expr.df_test
    target_root = util.config_val("EXPORT", "root", "./exported_data/")
    orig_root = util.config_val("EXPORT", "orig_root", None)
    data_name = util.config_val("EXPORT", "data_name", "export")
    segments_as_files = eval(util.config_val("EXPORT", "segments_as_files", "False"))
    audeer.mkdir(target_root)
    splits = {"train": df_train, "test": df_test}
    df_all = pd.DataFrame()
    for split in splits:
        files = []
        df = splits[split]
        for idx, (file, start, end) in enumerate(df.index.to_list()):
            file_dir = os.path.dirname(file)
            if segments_as_files:
                signal, sampling_rate = audiofile.read(
                    file,
                    offset=start.total_seconds(),
                    duration=(end - start).total_seconds(),
                    always_2d=True,
                )
                file_name = (
                    os.path.splitext(file)[0] + "_" + start.total_seconds() + ".wav"
                )
                wav_folder = (
                    f"{target_root}/{os.path.basename(os.path.normpath(orig_root))}"
                )
                audeer.mkdir(wav_folder)
                new_rel_path = file_dir[
                    file_dir.index(orig_root) + 1 + len(orig_root) :
                ]
                new_file_path = f"{wav_folder}/{new_rel_path}"
                audeer.mkdir(new_file_path)
                new_file_name = f"{new_file_path}/{file_name}"
                audiofile.write(new_file_name, signal, sampling_rate)
                new_file_name = os.path.relpath(new_file_name, target_root)
                files.append(new_file_name)
            else:
                file_name = os.path.basename(file)
                wav_folder = (
                    f"{target_root}/{os.path.basename(os.path.normpath(orig_root))}"
                )
                audeer.mkdir(wav_folder)
                new_rel_path = file_dir[
                    file_dir.index(orig_root) + 1 + len(orig_root) :
                ]
                new_file_path = f"{wav_folder}/{new_rel_path}"
                audeer.mkdir(new_file_path)
                new_file_name = f"{new_file_path}/{file_name}"
                if not os.path.exists(new_file_name):
                    shutil.copyfile(file, new_file_name)
                new_file_name = os.path.relpath(new_file_name, target_root)
                files.append(new_file_name)
        df = df.set_index(df.index.set_levels(files, level="file"))
        df["split"] = split
        df_all = pd.concat([df_all, df])
    # remove encoded labels
    target = util.config_val("DATA", "target", "emotion")
    if "class_label" in df_all.columns:
        df_all = df_all.drop(columns=[target])
        df_all = df_all.rename(columns={"class_label": target})

    df_all.to_csv(f"{target_root}/{data_name}.csv")
    util.debug(f"saved {data_name}.csv to {target_root}, {df.shape[0]} samples.")

    print("DONE")


if __name__ == "__main__":
    main()



""""
A predictor for PESQ - Perceptual Evaluation of Speech Quality.
"""

import ast

import numpy as np

import nkululeko.glob_conf as glob_conf
from nkululeko.feature_extractor import FeatureExtractor
from nkululeko.utils.util import Util


class PESQPredictor:
    """
    PESQPredictor
    predicting PESQ

    """

    def __init__(self, df):
        self.df = df
        self.util = Util("pesqPredictor")

    def predict(self, split_selection):
        self.util.debug(f"estimating PESQ for {split_selection} samples")
        return_df = self.df.copy()
        feats_name = "_".join(ast.literal_eval(glob_conf.config["DATA"]["databases"]))
        self.feature_extractor = FeatureExtractor(
            self.df, ["squim"], feats_name, split_selection
        )
        result_df = self.feature_extractor.extract()
        # replace missing values by 0
        result_df = result_df.fillna(0)
        result_df = result_df.replace(np.nan, 0)
        result_df.replace([np.inf, -np.inf], 0, inplace=True)
        pred_vals = result_df.pesq * 100
        return_df["pesq_pred"] = pred_vals.astype("int") / 100
        return return_df



# demo_predictor.py
import os

import audformat
import audiofile
import numpy as np
import pandas as pd

import nkululeko.glob_conf as glob_conf
from nkululeko.utils.util import Util


class Demo_predictor:
    def __init__(self, model, file, is_list, feature_extractor, label_encoder, outfile):
        """Constructor setting up name and configuration."""
        self.model = model
        self.feature_extractor = feature_extractor
        self.label_encoder = label_encoder
        self.is_list = is_list
        self.sr = 16000
        self.target = glob_conf.config["DATA"]["target"]
        self.util = Util("demo_predictor")
        self.file = file
        self.outfile = outfile

    def run_demo(self):
        signal, sr = None, 0
        if self.file is not None:
            if not self.is_list:
                sig, sr = audiofile.read(self.file)
                print(
                    f"predicting file: {self.file}, len: {len(sig)} bytes,"
                    f" sampling rate: {sr}"
                )
                self.predict_signal(sig, sr)
            else:
                df_res = pd.DataFrame()
                # first, try to read the list as audformat and use the index
                file_list = []
                try:
                    in_df = audformat.utils.read_csv(self.file)
                    if audformat.is_segmented_index(in_df.index):
                        self.util.error(
                            f"segmented index not implemented yet: {self.file}"
                        )
                    else:
                        file_list = in_df.index.values
                except (ValueError, AttributeError):
                    with open(self.file) as f:
                        first = True
                        for index, line in enumerate(f):
                            # first line might be "file"
                            if self.file.endswith(".csv") and first:
                                first = False
                            else:
                                file_list.append(line)
                for file_name in file_list:
                    test_folder = glob_conf.config["DATA"]["test_folder"]
                    file_path = os.path.join(test_folder, file_name.strip())
                    sig, sr = audiofile.read(file_path)
                    print(f"predicting file {file_path}")
                    res_dict = self.predict_signal(sig, sr)
                    df_tmp = pd.DataFrame(res_dict, index=[file_name.strip()])
                    df_res = pd.concat([df_res, df_tmp], ignore_index=False)
                df_res = df_res.set_index(df_res.index.rename("file"))
                # save only filename and prediction (df_tmp) by default
                # drop other columns
                # df_res = df_res[["predicted"]]
                if self.outfile is not None:
                    df_res.to_csv(self.outfile)
                else:
                    self.util.debug(df_res)
        else:
            answer = input("want to record y/n?")
            while answer == "y":
                signal = self.record_audio(3)
                self.predict_signal(signal, self.sr)
                answer = input("want to record y/n?")

    #            self.play_audio(signal)

    def predict_signal(self, signal, sr):
        features = self.feature_extractor.extract_sample(signal, sr)
        scale_feats = self.util.config_val("FEATS", "scale", False)
        if scale_feats:
            # standard normalize the input features
            features = (features - features.mean()) / (features.std())
        features = np.nan_to_num(features)
        result_dict = self.model.predict_sample(features)
        dict_2 = {}
        if self.util.exp_is_classification():
            keys = result_dict.keys()
            if self.label_encoder is not None:
                for i, k in enumerate(keys):
                    ak = np.array(int(k)).reshape(1)
                    lab = self.label_encoder.inverse_transform(ak)[0]
                    dict_2[lab] = f"{result_dict[k]:.3f}"
                dict_2["predicted"] = max(dict_2, key=dict_2.get)
                print(dict_2)
                return dict_2
            else:
                print(result_dict)
                return result_dict
        else:
            # experiment is regression and returns one estimation
            dict_2["predicted"] = result_dict
            print(dict_2)
            return dict_2

    def record_audio(self, seconds):
        import sounddevice as sd

        print("recording ...", flush=True)
        y = sd.rec(int(seconds * self.sr), samplerate=self.sr, channels=1)
        sd.wait()
        y = y.T
        return y

    def play_audio(self, signal):
        import sounddevice as sd

        print("playback ...")
        sd.play(signal.T, self.sr)
        status = sd.wait()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# @filename: process_database.py
# @description: Process the ASED database 


import argparse
import shutil
from importlib.resources import path
from pathlib import Path

import pandas as pd
from joblib import delayed

emotion_map = {
    "a": "anger",
    "h": "happiness",
    "n": "neutral",
    "f": "fear",
    "s": "sadness",
}


def main():
    parser = argparse.ArgumentParser(
        usage="python3 process_database.py database output"
    )
    parser.add_argument("--data_dir", type=str, default="ASED_V1", help="Path to the ASED database")
    parser.add_argument("--out_dir", type=str, default=".", help="Path to the output directory")
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    out_dir = Path(args.out_dir)

    paths = list(data_dir.glob("**/*.wav"))
    emotions = [emotion_map[path.stem[0].lower()] for path in paths]
    genders = [["female", "male"][int(p.stem[9:11]) - 1] for p in paths]
    spekaers = [p.stem[-2:] for p in paths]
    languages = ["amharic" for p in paths]

    # convert to df
    df = pd.DataFrame({"file": paths, "emotion": emotions, "gender": genders, "speaker": spekaers, "language": languages})

    # allocate speakers >= 55 for test
    df_test = df[df["speaker"] > "55"]
    df_train = df.drop(df_test.index)


    # save to csv
    df_train.to_csv(out_dir / "ased_train.csv", index=False)
    df_test.to_csv(out_dir / "ased_test.csv", index=False)
    df.to_csv(out_dir / "ased.csv", index=False)


if __name__ == "__main__":
    main()
    


"""
latex_writer.py
print out report as latex file and pdf
"""

from pylatex import Command, Document, Figure, Section, Subsection
from pylatex.utils import NoEscape

from nkululeko.utils.util import Util


class LatexWriter:
    def __init__(self):
        self.util = Util("LatexWriter")
        title = self.util.config_val("REPORT", "title", "Nkululeko report")
        author = self.util.config_val("REPORT", "author", "anon")
        doc = Document()
        doc.preamble.append(Command("title", title))
        doc.preamble.append(Command("author", author))
        doc.preamble.append(Command("date", NoEscape(r"\today")))
        doc.append(NoEscape(r"\maketitle"))
        self.doc = doc

    def add_items_for_section(self, section, contents):
        with self.doc.create(Section(section)):
            for ri in contents:
                with self.doc.create(Subsection(ri.caption)):
                    self.doc.append(ri.contents)
                    if ri.has_image:
                        with self.doc.create(Figure(position="h!")) as pic:
                            pic.add_image(ri.image, width="250px")
                            pic.add_caption(ri.caption)
                            # reference = pic.dumps_as_content()
                            # self.doc.append(f"See figure: {reference}")

    def finish_doc(self):
        from subprocess import CalledProcessError

        target_filename = self.util.config_val("REPORT", "latex", "nkululeko_latex")
        target_dir = self.util.get_exp_dir()
        path = "/".join([target_dir, target_filename])
        self.util.debug(f"Generated latex report to {path}")
        try:
            self.doc.generate_pdf(path, clean_tex=False)
        except CalledProcessError as e:
            self.util.debug(f"Error while generating PDF file: {e}")



import argparse
import glob
import os
from os.path import basename
from os.path import join as path_join
from os.path import splitext

import pandas as pd

LABEL_DIR_PATH = 'Labels'
WAV_DIR_PATH = 'Audios'


def get_wav_paths_podcast(paths):
    wav_paths = glob.glob(f"{paths}/**/*.wav", recursive=True)
    wav_dict = {}
    for wav_path in wav_paths:
        wav_name = splitext(basename(wav_path))[0]
        wav_dict[wav_name] = wav_path
    return wav_dict

def preprocess(data_dirs, paths, wav_paths, out_path):
    meta_data = []
    for idx, row in paths.iterrows():
        if row['EmoClass'] not in ['A', 'H', 'N', 'S']:
            continue
        # remove outlier (stereo, 44.1 kHz)
        if row['FileName'] == 'MSP-PODCAST_1023_0235.wav':
            continue
        filename = row['FileName'][:-4]
        label = row['EmoClass']
        valence = row['EmoVal']
        arousal = row['EmoAct']
        dominance = row['EmoDom']
        #speaker = row['SpkrID']
        #gender = row['Gender']
        for r in (('A', 'ang'), ('H', 'hap'), 
                ('N', 'neu'), ('S', 'sad')):
            label = label.replace(*r)
        meta_data.append({
                    'file': wav_paths[filename],
                    'emotion': label,
                    'valence': valence,
                    'arousal': arousal,
                    'dominance': dominance,
                    #'speaker': speaker,
                    #'gender': gender
                    })

    # write to csv
    df = pd.DataFrame(meta_data)
    df.to_csv(out_path, index=False)


def main(data_dir):
    """Main function."""
    out_dir = os.path.dirname(__file__)
    wav_paths = get_wav_paths_podcast(path_join(f"{data_dir}/{WAV_DIR_PATH}"))
    data = pd.read_csv(path_join(f"{data_dir}/{LABEL_DIR_PATH}",
                                       'labels_concensus.csv'))
    # Train and test1 1 are used for test
    train_path = data.loc[
        (data.Split_Set == 'Train')]
    # Development/validation set
    dev_path = data.loc[
        (data['Split_Set'] == 'Validation')]
    # Using test1 as test
    test_path1 = data.loc[
        (data['Split_Set'] == 'Test1')]  
    test_path2 = data.loc[
        (data['Split_Set'] == 'Test2')]
    
    preprocess(data_dir, train_path, wav_paths, 
               path_join(f"{out_dir}", 'podcast_train.csv'))
    preprocess(data_dir, dev_path, wav_paths, path_join(f"{out_dir}",
                                                         'podcast_dev.csv'))
    preprocess(data_dir, test_path1, wav_paths, 
               path_join(f"{out_dir}", 'podcast_test1.csv'))
    preprocess(data_dir, test_path2, wav_paths,
                path_join(f"{out_dir}", 'podcast_test2.csv'))


if __name__ == "__main__":
    """input the argument when calling this file
    e.g. python PODCAST_preprocess /path/to/PODCAST"""
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str, default='./MSP-PODCAST-Publish-1.8', help='Path to the dataset')
    args = parser.parse_args()
    main(args.data_dir)



# test_pretrain.py
import argparse
import configparser
import json
import os.path

import audeer
import audiofile
import audmetric
import datasets
import numpy as np
import pandas as pd
import torch
import transformers

import nkululeko.experiment as exp
import nkululeko.glob_conf as glob_conf
import nkululeko.models.finetune_model as fm
from nkululeko.constants import VERSION
from nkululeko.utils.util import Util


def doit(config_file):
    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f"ERROR: no such file: {config_file}")
        exit()

    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)

    # create a new experiment
    expr = exp.Experiment(config)
    module = "test_pretrain"
    expr.set_module(module)
    util = Util(module)
    util.debug(
        f"running {expr.name} from config {config_file}, nkululeko version"
        f" {VERSION}"
    )

    if util.config_val("EXP", "no_warnings", False):
        import warnings

        warnings.filterwarnings("ignore")

    # load the data
    expr.load_datasets()

    # split into train and test
    expr.fill_train_and_tests()
    util.debug(f"train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}")

    model_root = util.get_path("model_dir")
    log_root = audeer.mkdir("log")
    torch_root = audeer.path(model_root, "torch")

    metrics_gender = {
        "UAR": audmetric.unweighted_average_recall,
        "ACC": audmetric.accuracy,
    }

    sampling_rate = 16000
    max_duration_sec = 8.0

    model_path = "facebook/wav2vec2-large-robust-ft-swbd-300h"
    num_layers = None

    os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
    os.environ["CUDA_VISIBLE_DEVICES"] = "0"

    batch_size = 16
    accumulation_steps = 4
    # create dataset

    dataset = {}
    target_name = glob_conf.target
    data_sources = {
        "train": pd.DataFrame(expr.df_train[target_name]),
        "dev": pd.DataFrame(expr.df_test[target_name]),
    }

    for split in ["train", "dev"]:
        df = data_sources[split]
        df[target_name] = df[target_name].astype("float")

        y = pd.Series(
            data=df.itertuples(index=False, name=None),
            index=df.index,
            dtype=object,
            name="labels",
        )

        y.name = "targets"
        df = y.reset_index()
        df.start = df.start.dt.total_seconds()
        df.end = df.end.dt.total_seconds()

        print(f"{split}: {len(df)}")

        ds = datasets.Dataset.from_pandas(df)
        dataset[split] = ds

    dataset = datasets.DatasetDict(dataset)

    # load pre-trained model
    le = glob_conf.label_encoder
    mapping = dict(zip(le.classes_, range(len(le.classes_))))
    target_mapping = {k: int(v) for k, v in mapping.items()}
    target_mapping_reverse = {value: key for key, value in target_mapping.items()}

    config = transformers.AutoConfig.from_pretrained(
        model_path,
        num_labels=len(target_mapping),
        label2id=target_mapping,
        id2label=target_mapping_reverse,
        finetuning_task=target_name,
    )
    if num_layers is not None:
        config.num_hidden_layers = num_layers
    setattr(config, "sampling_rate", sampling_rate)
    setattr(config, "data", util.get_data_name())

    vocab_dict = {}
    with open("vocab.json", "w") as vocab_file:
        json.dump(vocab_dict, vocab_file)
    tokenizer = transformers.Wav2Vec2CTCTokenizer("./vocab.json")
    tokenizer.save_pretrained(".")

    feature_extractor = transformers.Wav2Vec2FeatureExtractor(
        feature_size=1,
        sampling_rate=16000,
        padding_value=0.0,
        do_normalize=True,
        return_attention_mask=True,
    )
    processor = transformers.Wav2Vec2Processor(
        feature_extractor=feature_extractor,
        tokenizer=tokenizer,
    )
    assert processor.feature_extractor.sampling_rate == sampling_rate

    model = fm.Model.from_pretrained(
        model_path,
        config=config,
    )
    model.freeze_feature_extractor()
    model.train()

    # training

    def data_collator(data):

        files = [d["file"] for d in data]
        starts = [d["start"] for d in data]
        ends = [d["end"] for d in data]
        targets = [d["targets"] for d in data]

        signals = []
        for file, start, end in zip(
            files,
            starts,
            ends,
        ):
            offset = start
            duration = end - offset
            if max_duration_sec is not None:
                duration = min(duration, max_duration_sec)
            signal, _ = audiofile.read(
                file,
                offset=offset,
                duration=duration,
            )
            signals.append(signal.squeeze())

        input_values = processor(
            signals,
            sampling_rate=sampling_rate,
            padding=True,
        )
        batch = processor.pad(
            input_values,
            padding=True,
            return_tensors="pt",
        )

        batch["labels"] = torch.tensor(targets)

        return batch

    def compute_metrics(p: transformers.EvalPrediction):

        truth_gender = p.label_ids[:, 0].astype(int)
        preds = p.predictions
        preds_gender = np.argmax(preds, axis=1)

        scores = {}

        for name, metric in metrics_gender.items():
            scores[f"gender-{name}"] = metric(truth_gender, preds_gender)

        scores["combined"] = scores["gender-UAR"]

        return scores

    targets = pd.DataFrame(dataset["train"]["targets"])
    counts = targets[0].value_counts().sort_index()
    train_weights = 1 / counts
    train_weights /= train_weights.sum()

    print(train_weights)

    criterion_gender = torch.nn.CrossEntropyLoss(
        weight=torch.Tensor(train_weights).to("cuda"),
    )

    class Trainer(transformers.Trainer):

        def compute_loss(
            self,
            model,
            inputs,
            return_outputs=False,
        ):

            targets = inputs.pop("labels").squeeze()
            targets_gender = targets.type(torch.long)

            outputs = model(**inputs)
            logits_gender = outputs[0].squeeze()

            loss_gender = criterion_gender(logits_gender, targets_gender)

            loss = loss_gender

            return (loss, outputs) if return_outputs else loss

    num_steps = len(dataset["train"]) // (batch_size * accumulation_steps) // 5
    num_steps = max(1, num_steps)
    print(num_steps)

    training_args = transformers.TrainingArguments(
        output_dir=model_root,
        logging_dir=log_root,
        per_device_train_batch_size=batch_size,
        per_device_eval_batch_size=batch_size,
        gradient_accumulation_steps=accumulation_steps,
        evaluation_strategy="steps",
        num_train_epochs=5.0,
        fp16=True,
        save_steps=num_steps,
        eval_steps=num_steps,
        logging_steps=num_steps,
        learning_rate=1e-4,
        save_total_limit=2,
        metric_for_best_model="combined",
        greater_is_better=True,
        load_best_model_at_end=True,
        remove_unused_columns=False,
        report_to="none",
    )

    trainer = Trainer(
        model=model,
        data_collator=data_collator,
        args=training_args,
        compute_metrics=compute_metrics,
        train_dataset=dataset["train"],
        eval_dataset=dataset["dev"],
        tokenizer=processor.feature_extractor,
        callbacks=[transformers.integrations.TensorBoardCallback()],
    )
    if False:
        trainer.train()
        trainer.save_model(torch_root)

    modelnew = fm.Model.from_pretrained(
        torch_root,
        config=config,
    )
    print(f"loaded new model type{type(modelnew)}")
    import audiofile

    signal, _ = audiofile.read("./test.wav", always_2d=True)
    result = modelnew.predict(signal)
    print(result)

    print("DONE")


def main(src_dir):
    parser = argparse.ArgumentParser(description="Call the nkululeko framework.")
    parser.add_argument("--config", default="exp.ini", help="The base configuration")
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config
    else:
        config_file = f"{src_dir}/exp.ini"
    doit(config_file)


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd)  # use this if you want to state the config file path on command line



# feats_ast.py
import os

import numpy as np
import pandas as pd
import torch
import torchaudio
from tqdm import tqdm
from transformers import ASTModel, AutoProcessor

import nkululeko.glob_conf as glob_conf
from nkululeko.feat_extract.featureset import Featureset


class Ast(Featureset):
    """Class to extract AST (Audio Spectrogram Transformer) embeddings"""

    def __init__(self, name, data_df, feat_type):
        super().__init__(name, data_df, feat_type)
        cuda = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = self.util.config_val("MODEL", "device", cuda)
        self.model_initialized = False
        self.feat_type = feat_type

    def init_model(self):
        self.util.debug("loading AST model...")
        model_path = self.util.config_val(
            "FEATS", "ast.model", "MIT/ast-finetuned-audioset-10-10-0.4593"
        )
        self.processor = AutoProcessor.from_pretrained(model_path)
        self.model = ASTModel.from_pretrained(model_path).to(self.device)
        print(f"initialized AST model on {self.device}")
        self.model.eval()
        self.model_initialized = True

    def extract(self):
        """Extract the features or load them from disk if present."""
        store = self.util.get_path("store")
        storage = f"{store}{self.name}.pkl"
        extract = self.util.config_val("FEATS", "needs_feature_extraction", False)
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if extract or no_reuse or not os.path.isfile(storage):
            if not self.model_initialized:
                self.init_model()
            self.util.debug("extracting wavlm embeddings, this might take a while...")
            emb_series = pd.Series(index=self.data_df.index, dtype=object)
            length = len(self.data_df.index)
            for idx, (file, start, end) in enumerate(
                tqdm(self.data_df.index.to_list())
            ):
                signal, sampling_rate = torchaudio.load(
                    file,
                    frame_offset=int(start.total_seconds() * 16000),
                    num_frames=int((end - start).total_seconds() * 16000),
                )
                # make mono if stereo
                if signal.shape[0] == 2:
                    signal = torch.mean(signal, dim=0, keepdim=True)

                assert (
                    sampling_rate == 16000
                ), f"sampling rate should be 16000 but is {sampling_rate}"
                emb = self.get_embeddings(signal, sampling_rate, file)
                emb_series.iloc[idx] = emb
            self.df = pd.DataFrame(emb_series.values.tolist(), index=self.data_df.index)
            self.df.to_pickle(storage)
            try:
                glob_conf.config["DATA"]["needs_feature_extraction"] = "false"
            except KeyError:
                pass
        else:
            self.util.debug(f"reusing extracted {self.feat_type} embeddings")
            self.df = pd.read_pickle(storage)
            if self.df.isnull().values.any():
                # nanrows = self.df.columns[self.df.isna().any()].tolist()
                # print(nanrows)
                self.util.error(
                    f"got nan: {self.df.shape} {self.df.isnull().sum().sum()}"
                )

    def get_embeddings(self, signal, sampling_rate, file):
        """Extract embeddings from raw audio signal."""
        try:
            inputs = self.processor(
                signal.numpy(), sampling_rate=sampling_rate, return_tensors="pt"
            )

            inputs = {k: v.to(self.device) for k, v in inputs.items()}

            with torch.no_grad():
                # Get the hidden states
                outputs = self.model(**inputs)

            # Get the hidden states from the last layer
            last_hidden_state = outputs.last_hidden_state

            # print(f"last_hidden_state shape: {last_hidden_state.shape}")
            # Average pooling over the time dimension
            embeddings = torch.mean(last_hidden_state, dim=1)
            embeddings = embeddings.cpu().numpy()

            # print(f"hs shape: {embeddings.shape}")
            # hs shape: (1, 768)

        except Exception as e:
            self.util.error(
                f"Error extracting embeddings for file {file}: {str(e)}, fill with"
            )
            return np.zeros(
                self.model.config.hidden_size
            )  # Return zero vector on error
        return embeddings.ravel()

    def extract_sample(self, signal, sr):
        self.init_model()
        feats = self.get_embeddings(signal, sr, "no file")
        return feats



# from nkululeko.constants import VERSION
from setuptools import setup

setup(
    use_scm_version=True,
    # version=VERSION,
    entry_points={
        "console_scripts": [
            # 'nkululeko=nkululeko.nkululeko:main',
            "nkululeko.explore=nkululeko.explore:main",
            "nkululeko.nkululeko=nkululeko.nkululeko:main",
            "nkululeko.augment=nkululeko.augment:main",
            "nkululeko.demo=nkululeko.demo:main",
            "nkululeko.export=nkululeko.export:main",
            "nkululeko.predict=nkululeko.predict:main",
            "nkululeko.resample=nkululeko.resample:main",
            "nkululeko.segment=nkululeko.segment:main",
            "nkululeko.test=nkululeko.test:main",
        ],
    },
)



# plots.py
import ast

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy import stats
from sklearn.manifold import TSNE

import nkululeko.glob_conf as glob_conf
import nkululeko.utils.stats as su
from nkululeko.reporting.defines import Header
from nkululeko.reporting.report_item import ReportItem
from nkululeko.utils.util import Util


class Plots:
    def __init__(self):
        """Initializing the util system."""
        self.util = Util("plots")
        self.format = self.util.config_val("PLOT", "format", "png")
        self.target = self.util.config_val("DATA", "target", "emotion")

    def plot_distributions_speaker(self, df):
        df_speakers = pd.DataFrame()
        pd.options.mode.chained_assignment = None  # default='warn'
        for s in df.speaker.unique():
            df_speaker = df[df.speaker == s]
            df_speaker["samplenum"] = df_speaker.shape[0]
            df_speakers = pd.concat([df_speakers, df_speaker.head(1)])
        # plot the distribution of samples per speaker
        # one up because of the runs
        fig_dir = self.util.get_path("fig_dir") + "../"
        self.util.debug("plotting samples per speaker")
        if "gender" in df_speakers:
            filename = "samples_value_counts"
            ax = (
                df_speakers.groupby("samplenum")["gender"]
                .value_counts()
                .unstack()
                .plot(
                    kind="bar",
                    stacked=True,
                    title=f"samples per speaker ({df_speakers.shape[0]})",
                    rot=0,
                )
            )
            ax.set_ylabel("number of speakers")
            ax.set_xlabel("number of samples")
            self.save_plot(
                ax,
                "Samples per speaker",
                f"Samples per speaker ({df_speakers.shape[0]})",
                filename,
                "speakers",
            )

            # fig.clear()
        else:
            filename = "samples_value_counts"
            ax = (
                df_speakers["samplenum"]
                .value_counts()
                .sort_values()
                .plot(
                    kind="bar",
                    stacked=True,
                    title=f"samples per speaker ({df_speakers.shape[0]})",
                    rot=0,
                )
            )
            ax.set_ylabel("number of speakers")
            ax.set_xlabel("number of samples")
            self.save_plot(
                ax,
                "Sample value counts",
                f"Samples per speaker ({df_speakers.shape[0]})",
                filename,
                "speakers",
            )

        self.plot_distributions(df_speakers, type_s="speakers")

    def plot_distributions(self, df, type_s="samples"):
        class_label, df = self._check_binning("class_label", df)
        attributes = ast.literal_eval(
            self.util.config_val("EXPL", "value_counts", False)
        )
        # always plot the distribution of the main attribute
        filename = f"{class_label}_distribution"
        if self.util.is_categorical(df[class_label]):
            ax = df[class_label].value_counts().plot(kind="bar")
        else:
            # for continuous variables, also add a discretized version
            binned_data = self.util.continuous_to_categorical(df[class_label])
            ax = binned_data.value_counts().plot(kind="bar")
            filename_binned = f"{class_label}_discreet"
            self.save_plot(
                ax,
                "Sample value counts",
                filename_binned,
                filename_binned,
                type_s,
            )
            dist_type = self.util.config_val("EXPL", "dist_type", "hist")
            ax = df[class_label].plot(kind=dist_type)

        self.save_plot(
            ax,
            "Sample value counts",
            filename,
            filename,
            type_s,
        )

        for att in attributes:
            if len(att) == 1:
                att1 = att[0]
                if att1 == self.target:
                    self.util.debug(f"no need to correlate {att1} with itself")
                    return
                if att1 not in df:
                    self.util.error(f"unknown feature: {att1}")
                att1, df = self._check_binning(att1, df)
                self.util.debug(f"plotting {att1}")
                filename = f"{self.target}-{att1}"
                if self.util.is_categorical(df[class_label]):
                    if self.util.is_categorical(df[att1]):
                        ax, caption = self._plot2cat(
                            df, class_label, att1, self.target, type_s
                        )
                    else:
                        ax, caption = self.plotcatcont(
                            df, class_label, att1, att1, type_s
                        )
                else:
                    if self.util.is_categorical(df[att1]):
                        ax, caption = self.plotcatcont(
                            df, att1, class_label, att1, type_s
                        )
                    else:
                        ax, caption = self._plot2cont(df, class_label, att1, type_s)
                self.save_plot(
                    ax,
                    caption,
                    f"Correlation of {self.target} and {att[0]}",
                    filename,
                    type_s,
                )
                # fig.clear()           # avoid error
            elif len(att) == 2:
                att1 = att[0]
                att2 = att[1]
                if att1 == self.target or att2 == self.target:
                    self.util.debug(f"no need to correlate {self.target} with itself")
                    return
                if att1 not in df:
                    self.util.error(f"unknown feature: {att1}")
                if att2 not in df:
                    self.util.error(f"unknown feature: {att2}")
                att1, df = self._check_binning(att1, df)
                att2, df = self._check_binning(att2, df)
                self.util.debug(f"plotting {att}")
                filename = f"{att1}-{att2}"
                filename = f"{self.target}-{filename}"
                if self.util.is_categorical(df["class_label"]):
                    if self.util.is_categorical(df[att1]):
                        if self.util.is_categorical(df[att2]):
                            # class_label = cat, att1 = cat, att2 = cat
                            ax, caption = self._plot2cat(df, att1, att2, att1, type_s)
                        else:
                            # class_label = cat, att1 = cat, att2 = cont
                            ax, caption = self.plotcatcont(df, att1, att2, att1, type_s)
                    else:
                        if self.util.is_categorical(df[att2]):
                            # class_label = cat, att1 = cont, att2 = cat
                            ax, caption = self.plotcatcont(df, att2, att1, att2, type_s)
                        else:
                            # class_label = cat, att1 = cont, att2 = cont
                            ax, caption = self._plot2cont_cat(
                                df, att1, att2, "class_label", type_s
                            )
                else:  # class_label is continuous
                    if self.util.is_categorical(df[att1]):
                        if self.util.is_categorical(df[att2]):
                            # class_label = cont, att1 = cat, att2 = cat
                            ax, caption = self._plot2cat(df, att1, att2, att1, type_s)
                        else:
                            # class_label = cont, att1 = cat, att2 = cont
                            ax, caption = self._plot2cont_cat(
                                df, att2, "class_label", att1, type_s
                            )
                    else:
                        if self.util.is_categorical(df[att2]):
                            # class_label = cont, att1 = cont, att2 = cat
                            ax, caption = self._plot2cont_cat(
                                df, att1, "class_label", att2, type_s
                            )
                        else:
                            # class_label = cont, att1 = cont, att2 = cont
                            ax, caption = self._plot2cont(df, att1, att2, type_s)

                self.save_plot(
                    ax, caption, f"Correlation of {att1} and {att2}", filename, type_s
                )

            else:
                self.util.error(
                    "plot value counts: the plot distribution descriptor for"
                    f" {att} has more than 2 values. Perhaps you forgot to state a list of lists?"
                )

    def save_plot(self, ax, caption, header, filename, type_s):
        # one up because of the runs
        fig_dir = self.util.get_path("fig_dir") + "../"
        fig_plots = ax.figure
        # avoid warning
        # plt.tight_layout()
        img_path = f"{fig_dir}{filename}_{type_s}.{self.format}"
        plt.savefig(img_path)
        plt.close(fig_plots)
        self.util.debug(f"Saved plot to {img_path}")
        glob_conf.report.add_item(
            ReportItem(
                Header.HEADER_EXPLORE,
                header,
                caption,
                img_path,
            )
        )

    def _check_binning(self, att, df):
        bin_reals_att = eval(self.util.config_val("EXPL", f"{att}.bin_reals", "False"))
        if bin_reals_att:
            self.util.debug(f"binning continuous variable {att} to categories")
            att_new = f"{att}_binned"
            df[att_new] = self.util.continuous_to_categorical(df[att]).values
            att = att_new
        return att, df

    def _plot2cont_cat(self, df, cont1, cont2, cat, ylab):
        """Plot relation of two continuous distributions with one categorical."""
        pearson = stats.pearsonr(df[cont1], df[cont2])
        # trunc to three digits
        pearson = int(pearson[0] * 1000) / 1000
        pearson_string = f"PCC: {pearson}"
        ax = sns.lmplot(data=df, x=cont1, y=cont2, hue=cat)
        caption = f"{ylab} {df.shape[0]}. {pearson_string}"
        ax.figure.suptitle(caption)
        return ax, caption

    def _plot2cont(self, df, col1, col2, ylab):
        """Plot relation of two continuous distributions."""
        pearson = stats.pearsonr(df[col1], df[col2])
        # trunc to three digits
        pearson = int(pearson[0] * 1000) / 1000
        pearson_string = f"PCC: {pearson}"
        ax = sns.lmplot(data=df, x=col1, y=col2)
        caption = f"{ylab} {df.shape[0]}. {pearson_string}"
        ax.figure.suptitle(caption)
        return ax, caption

    def plotcatcont(self, df, cat_col, cont_col, xlab, ylab):
        """Plot relation of categorical distribution with continuous."""
        dist_type = self.util.config_val("EXPL", "dist_type", "kde")
        cats, cat_str, es = su.get_effect_size(df, cat_col, cont_col)
        model_type = self.util.get_model_type()
        if dist_type == "hist" and model_type != "tree":
            ax = sns.histplot(df, x=cont_col, hue=cat_col, kde=True)
            caption = f"{ylab} {df.shape[0]}. {cat_str} ({cats}):" f" {es}"
            ax.set_title(caption)
            ax.set_xlabel(f"{cont_col}")
            ax.set_ylabel(f"number of {ylab}")
        else:
            ax = sns.displot(
                df, x=cont_col, hue=cat_col, kind="kde", fill=True, warn_singular=False
            )
            ax.set(xlabel=f"{cont_col}")
            caption = f"{ylab} {df.shape[0]}. {cat_str} ({cats}):" f" {es}"
            ax.figure.suptitle(caption)
        return ax, caption

    def _plot2cat(self, df, col1, col2, xlab, ylab):
        """Plot relation of 2 categorical distributions."""
        crosstab = pd.crosstab(index=df[col1], columns=df[col2])
        res_pval = stats.chi2_contingency(crosstab)
        res_pval = int(res_pval[1] * 1000) / 1000
        caption = f"{ylab} {df.shape[0]}. P-val chi2: {res_pval}"
        ax = (
            df.groupby(col1, observed=False)[col2]
            .value_counts()
            .unstack()
            .plot(kind="bar", stacked=True, title=caption, rot=0)
        )
        ax.set_ylabel(f"number of {ylab}")
        ax.set_xlabel(xlab)
        return ax, caption

    def plot_durations(self, df, filename, sample_selection, caption=""):
        # one up because of the runs
        fig_dir = self.util.get_path("fig_dir") + "../"
        try:
            ax = sns.histplot(df, x="duration", hue="class_label", kde=True)
        except AttributeError as ae:
            self.util.warn(ae)
            ax = sns.histplot(df, x="duration", kde=True)
        min = self.util.to_3_digits(df.duration.min())
        max = self.util.to_3_digits(df.duration.max())
        title = f"Duration distr. for {sample_selection} {df.shape[0]}. min={min}, max={max}"
        ax.set_title(title)
        ax.set_xlabel("duration")
        ax.set_ylabel("number of samples")
        fig = ax.figure
        # plt.tight_layout()
        img_path = f"{fig_dir}{filename}_{sample_selection}.{self.format}"
        plt.savefig(img_path)
        plt.close(fig)
        glob_conf.report.add_item(
            ReportItem(
                Header.HEADER_EXPLORE,
                caption,
                title,
                img_path,
            )
        )

    def describe_df(self, name, df, target, filename):
        """Make a stacked barplot of samples and speakers per sex and target values. speaker, gender and target columns must be present"""
        fig_dir = self.util.get_path("fig_dir") + "../"  # one up because of the runs
        sampl_num = df.shape[0]
        sex_col = "gender"
        if target == "gender":
            sex_col = "class_label"
        if self.util.exp_is_classification() and target != "gender":
            target = "class_label"
        if df.is_labeled:
            if df.got_gender and df.got_speaker:
                spkr_num = df.speaker.nunique()
                female_smpl_num = df[df[sex_col] == "female"].shape[0]
                male_smpl_num = df[df[sex_col] == "male"].shape[0]
                self.util.debug(
                    f"plotting {name}: # samples: {sampl_num} (f:"
                    f" {female_smpl_num}, m: "
                    + f"{male_smpl_num}), # speakers: {spkr_num}"
                )
                # fig, axes = plt.subplots(nrows=1, ncols=2)
                fig, axes = plt.subplots(nrows=1, ncols=1)
                # df.groupby(target)['gender'].value_counts().unstack().plot(kind='bar', stacked=True, ax=axes[0], \
                #     title=f'samples ({sampl_num})')
                df.groupby(target)["gender"].value_counts().unstack().plot(
                    kind="bar", stacked=True, title=f"samples ({sampl_num})"
                )
                # df.groupby(target)['speaker'].nunique().plot(kind='bar', ax=axes[1], title=f'speakers ({spkr_num})')
            else:
                self.util.debug(f"plotting {name}: # samples: {sampl_num}")
                fig, axes = plt.subplots(nrows=1, ncols=1)
                df[target].value_counts().plot(
                    kind="bar", ax=axes, title=f"samples ({sampl_num})"
                )
            # plt.tight_layout()
            img_path = f"{fig_dir}{filename}.{self.format}"
            plt.savefig(img_path)
            fig.clear()
            plt.close(fig)
            glob_conf.report.add_item(
                ReportItem(
                    Header.HEADER_EXPLORE,
                    f"Overview on {df.shape[0]} samples",
                    "",
                    img_path,
                )
            )

    def scatter_plot(self, feats, label_df, label, dimred_type):
        dim_num = int(self.util.config_val("EXPL", "scatter.dim", 2))
        # one up because of the runs
        fig_dir = self.util.get_path("fig_dir") + "../"
        sample_selection = self.util.config_val("EXPL", "sample_selection", "all")
        filename = f"{label}_{self.util.get_feattype_name()}_{sample_selection}_{dimred_type}_{str(dim_num)}d"
        filename = f"{fig_dir}{filename}.{self.format}"
        self.util.debug(f"computing {dimred_type}, this might take a while...")
        data = None
        labels = label_df[label]
        if dimred_type == "tsne":
            data = self.getTsne(feats, dim_num)
        else:
            if dimred_type == "umap":
                import umap

                y = umap.UMAP(
                    n_neighbors=10,
                    random_state=0,
                    n_components=dim_num,
                ).fit_transform(feats.values)
            elif dimred_type == "pca":
                from sklearn.decomposition import PCA
                from sklearn.preprocessing import StandardScaler

                scaler = StandardScaler()
                pca = PCA(n_components=dim_num)
                y = pca.fit_transform(scaler.fit_transform(feats.values))
            else:
                self.util.error(
                    f"no such dimensionality reduction function: {dimred_type}"
                )
            if dim_num == 2:
                columns = ["Dim_1", "Dim_2"]
            elif dim_num == 3:
                columns = ["Dim_1", "Dim_2", "Dim_3"]
            else:
                self.util.error(f"wrong dimension number: {dim_num}")
            data = pd.DataFrame(
                y,
                feats.index,
                columns=columns,
            )

        if dim_num == 2:
            plot_data = np.vstack((data.T, labels)).T
            plot_df = pd.DataFrame(data=plot_data, columns=("Dim_1", "Dim_2", "label"))
            # plt.tight_layout()
            ax = (
                sns.FacetGrid(plot_df, hue="label", height=6)
                .map(plt.scatter, "Dim_1", "Dim_2")
                .add_legend()
            )
        elif dim_num == 3:
            from mpl_toolkits.mplot3d import Axes3D
            from sklearn.preprocessing import LabelEncoder

            le = LabelEncoder()

            labels_e = le.fit_transform(labels)
            plot_data = np.vstack((data.T, labels_e)).T
            plot_df = pd.DataFrame(
                data=plot_data, columns=("Dim_1", "Dim_2", "Dim_3", "label")
            )
            # plt.tight_layout()
            # axes instance
            fig = plt.figure(figsize=(6, 6))
            ax = Axes3D(fig, auto_add_to_figure=False)
            fig.add_axes(ax)
            # get colormap from seaborn
            # cmap = ListedColormap(sns.color_palette("hsv", 256).as_hex())
            color_dict = {
                0: "red",
                1: "blue",
                2: "green",
                3: "yellow",
                4: "purple",
                5: "#ff69b4",
                6: "black",
                7: "cyan",
                8: "magenta",
                9: "#faebd7",
                10: "#2e8b57",
                11: "#eeefff",
                12: "#da70d6",
                13: "#ff7f50",
                14: "#cd853f",
                15: "#bc8f8f",
                16: "#5f9ea0",
                17: "#daa520",
            }
            # plot
            # make the numbers bigger so they can be used as distinguishable colors
            labels_ex = [color_dict[xi] for xi in labels_e]
            sc = ax.scatter(
                plot_df.Dim_1,
                plot_df.Dim_2,
                plot_df.Dim_3,
                s=40,
                c=labels_ex,
                marker="o",
                # cmap=cmap,
                alpha=1,
            )
            ax.set_xlabel("Dim_1")
            ax.set_ylabel("Dim_2")
            ax.set_zlabel("Dim_3")
            # legend
            plt.legend(*sc.legend_elements(), bbox_to_anchor=(1.05, 1), loc=2)
        else:
            self.util.error(f"wrong dimension number: {dim_num}")
        fig = ax.figure
        plt.savefig(filename)
        fig.clear()
        plt.close(fig)
        glob_conf.report.add_item(
            ReportItem(
                Header.HEADER_EXPLORE,
                "Scatter plot",
                f"using {dimred_type}",
                filename,
            )
        )

    def getTsne(self, feats, dim_num, perplexity=30, learning_rate=200):
        """Make a TSNE plot to see whether features are useful for classification"""
        model = TSNE(
            n_components=dim_num,
            random_state=0,
            perplexity=perplexity,
            learning_rate=learning_rate,
        )
        tsne_data = model.fit_transform(feats)
        return tsne_data

    def plot_feature(self, title, feature, label, df_labels, df_features):
        # remove fullstops in the name
        feature_name = feature.replace(".", "-")
        # one up because of the runs
        fig_dir = self.util.get_path("fig_dir") + "../"
        filename = f"{fig_dir}feat_dist_{title}_{feature_name}.{self.format}"
        if self.util.is_categorical(df_labels[label]):
            df_plot = pd.DataFrame(
                {label: df_labels[label], feature: df_features[feature]}
            )
            ax = sns.violinplot(data=df_plot, x=label, y=feature)
            label = self.util.config_val("DATA", "target", "class_label")
            ax.set(title=f"{title} samples", xlabel=label)
        else:
            plot_df = pd.concat([df_labels, df_features], axis=1)
            ax, caption = self._plot2cont(plot_df, label, feature, feature)
        # def _plot2cont(self, df, col1, col2, xlab, ylab):

        fig = ax.figure
        # plt.tight_layout()
        plt.savefig(filename)
        fig.clear()
        plt.close(fig)
        caption = f"Feature plot for feature {feature}"
        content = caption
        glob_conf.report.add_item(
            ReportItem(
                Header.HEADER_EXPLORE,
                caption,
                content,
                filename,
            )
        )

    def plot_tree(self, model, features):
        from sklearn import tree

        ax = plt.gca()
        ax.figure.set_size_inches(100, 60)
        #        tree.plot_tree(model, ax = ax)
        tree.plot_tree(model, feature_names=list(features.columns), ax=ax)
        # plt.tight_layout()
        # print(ax)
        # one up because of the runs
        fig_dir = self.util.get_path("fig_dir") + "../"
        exp_name = self.util.get_exp_name(only_data=True)
        format = self.util.config_val("PLOT", "format", "png")
        filename = f"{fig_dir}{exp_name}EXPL_tree-plot.{format}"
        fig = ax.figure
        fig.savefig(filename)
        fig.clear()
        plt.close(fig)
        glob_conf.report.add_item(
            ReportItem(
                Header.HEADER_EXPLORE,
                "Tree plot",
                "for feature importance",
                filename,
            )
        )



# dataset.py
import ast
import os
import os.path
from random import sample

import audformat
import numpy as np
import pandas as pd

import nkululeko.glob_conf as glob_conf
from nkululeko.filter_data import DataFilter
from nkululeko.plots import Plots
from nkululeko.reporting.report_item import ReportItem
from nkululeko.utils.util import Util


class Dataset:
    """Class to represent datasets"""

    name = ""  # An identifier for the dataset
    config = None  # The configuration
    db = None  # The database object
    df = None  # The whole dataframe
    df_train = None  # The training split
    df_test = None  # The evaluation split

    def __init__(self, name):
        """Constructor setting up name and configuration"""
        self.name = name
        self.util = Util("dataset")
        self.target = self.util.config_val("DATA", "target", "none")
        self.plot = Plots()
        self.limit = int(self.util.config_val_data(self.name, "limit", 0))
        self.start_fresh = eval(self.util.config_val("DATA", "no_reuse", "False"))
        self.is_labeled, self.got_speaker, self.got_gender, self.got_age = (
            False,
            False,
            False,
            False,
        )

    def _get_tables(self):
        tables = []
        targets = self.util.config_val_data(self.name, "target_tables", False)
        if targets:
            target_tables = ast.literal_eval(targets)
            tables += target_tables
        files = self.util.config_val_data(self.name, "files_tables", False)
        if files:
            files_tables = ast.literal_eval(files)
            tables += files_tables
        tests = self.util.config_val_data(self.name, "test_tables", False)
        if tests:
            test_tables = ast.literal_eval(tests)
            tables += test_tables
        trains = self.util.config_val_data(self.name, "train_tables", False)
        if trains:
            train_tables = ast.literal_eval(trains)
            tables += train_tables
        return tables

    def _load_db(self):
        root = self.util.config_val_data(self.name, "", "")
        self.util.debug(f"{self.name}: loading from {root}")
        try:
            self.db = audformat.Database.load(root)
        except FileNotFoundError:
            self.util.error(f"{self.name}: no audformat database found at {root}")
        return root

    def _check_cols(self, df):
        rename_cols = self.util.config_val_data(self.name, "colnames", False)
        if rename_cols:
            col_dict = ast.literal_eval(rename_cols)
            df = df.rename(columns=col_dict)
            self.util.debug(f"renamed data columns: {col_dict}")
        return df

    def _report_load(self):
        speaker_num = 0
        if self.got_speaker:
            speaker_num = self.df.speaker.nunique()
        r_string = (
            f"{self.name}: loaded with {self.df.shape[0]} samples: got targets:"
            f" {self.is_labeled}, got speakers:"
            f" {self.got_speaker} ({speaker_num}), got sexes: {self.got_gender}"
        )
        self.util.debug(r_string)
        if glob_conf.report.initial:
            glob_conf.report.add_item(ReportItem("Data", "Load report", r_string))
            glob_conf.report.initial = False

    def load(self):
        """Load the dataframe with files, speakers and task labels"""
        # store the dataframe
        store = self.util.get_path("store")
        store_format = self.util.config_val("FEATS", "store_format", "pkl")
        store_file = f"{store}{self.name}.{store_format}"
        self.root = self._load_db()
        if not self.start_fresh and os.path.isfile(store_file):
            self.util.debug(f"{self.name}: reusing previously stored file {store_file}")
            self.df = self.util.get_store(store_file, store_format)
            self.is_labeled = self.target in self.df
            self.got_gender = "gender" in self.df
            self.got_age = "age" in self.df
            self.got_speaker = "speaker" in self.df
            self.util.copy_flags(self, self.df)
            self._report_load()
            return
        tables = self._get_tables()
        self.util.debug(f"{self.name}: loading tables: {tables}")
        # db = audb.load(root, )
        # map the audio file paths
        self.db.map_files(lambda x: os.path.join(self.root, x))
        # the dataframes (potentially more than one) with at least the file names
        df_files = self.util.config_val_data(self.name, "files_tables", "['files']")
        df_files_tables = ast.literal_eval(df_files)
        # The label for the target column
        self.col_label = self.util.config_val_data(self.name, "label", self.target)
        (
            df,
            self.is_labeled,
            self.got_speaker,
            self.got_gender,
            self.got_age,
        ) = self._get_df_for_lists(self.db, df_files_tables)
        if df.shape[0] > 0 and self.target == "none":
            self.df = df
            return
        if False in {
            self.is_labeled,
            self.got_speaker,
            self.got_gender,
            self.got_age,
        }:
            try:
                # There might be a separate table with the targets, e.g. emotion or age
                df_targets = self.util.config_val_data(
                    self.name, "target_tables", f"['{self.target}']"
                )
                df_target_tables = ast.literal_eval(df_targets)
                (
                    df_target,
                    got_target2,
                    got_speaker2,
                    got_gender2,
                    got_age2,
                ) = self._get_df_for_lists(self.db, df_target_tables)
                self.is_labeled = got_target2 or self.is_labeled
                self.got_speaker = got_speaker2 or self.got_speaker
                self.got_gender = got_gender2 or self.got_gender
                self.got_age = got_age2 or self.got_age
                if audformat.is_filewise_index(df_target.index):
                    try:
                        df_target = df_target.loc[df.index.get_level_values("file")]
                        df_target = df_target.set_index(df.index)
                    except KeyError:
                        # just a try...
                        pass
                if got_target2:
                    df[self.target] = df_target[self.target]
                if got_speaker2:
                    df["speaker"] = df_target["speaker"]
                if got_gender2:
                    df["gender"] = df_target["gender"]
                if got_age2:
                    df["age"] = df_target["age"].astype(int)
                # copy other column
                for column in df_target.columns:
                    if column not in [self.target, "age", "speaker", "gender"]:
                        df[column] = df_target[column]
            except audformat.core.errors.BadKeyError:
                if not self.is_labeled:
                    self.util.error(
                        f"Giving up: no target ({self.target}) column found"
                    )

        if self.is_labeled:
            # remember the target in case they get labelencoded later
            df["class_label"] = df[self.target]

        self.df = df
        self._report_load()

    def prepare(self):
        # ensure segmented index
        self.df = self.util.make_segmented_index(self.df)
        self.util.copy_flags(self, self.df)
        # check the type of numeric targets
        if not self.util.exp_is_classification():
            self.df[self.target] = self.df[self.target].astype(float)
        # add duration
        if "duration" not in self.df:
            start = self.df.index.get_level_values(1)
            end = self.df.index.get_level_values(2)
            self.df["duration"] = (end - start).total_seconds()
        elif self.df.duration.dtype == "timedelta64[ns]":
            self.df["duration"] = self.df["duration"].map(lambda x: x.total_seconds())
        # Perform some filtering if desired
        required = self.util.config_val_data(self.name, "required", False)
        if required:
            pre = self.df.shape[0]
            self.df = self.df[self.df[required].notna()]
            post = self.df.shape[0]
            self.util.debug(
                f"{self.name}: kept {post} samples with {required} (from {pre},"
                f" filtered {pre-post})"
            )

        datafilter = DataFilter(self.df)
        self.df = datafilter.all_filters(data_name=self.name)

        if self.got_speaker and self.util.config_val_data(
            self.name, "rename_speakers", False
        ):
            # we might need to append the database name to all speakers in case other databases have the same speaker names
            self.df.speaker = self.df.speaker.astype(str)
            self.df.speaker = self.df.speaker.apply(lambda x: self.name + x)

        # check if the target variable should be reversed
        def reverse_array(d, max):
            d = np.array(d)
            res = []
            for n in d:
                res.append(abs(n - max))
            return res

        reverse = eval(self.util.config_val_data(self.name, "reverse", "False"))
        if reverse:
            max = eval(self.util.config_val_data(self.name, "reverse.max", "False"))
            if max:
                max = float(max)
            else:
                max = self.df[self.target].values.max()
            self.util.debug(f"reversing target numbers with max values: {max}")
            self.df[self.target] = reverse_array(self.df[self.target].values, max)

        # check if the target variable should be scaled (z-transformed)
        scale = self.util.config_val_data(self.name, "scale", False)
        if scale:
            from sklearn.preprocessing import StandardScaler

            self.util.debug("scaling target variable to normal distribution")
            scaler = StandardScaler()
            self.df[self.target] = scaler.fit_transform(
                self.df[self.target].values.reshape(-1, 1)
            )

        # store the dataframe
        store = self.util.get_path("store")
        store_format = self.util.config_val("FEATS", "store_format", "pkl")
        store_file = f"{store}{self.name}.{store_format}"
        self.util.write_store(self.df, store_file, store_format)

    def _get_df_for_lists(self, db, df_files):
        is_labeled, got_speaker, got_gender, got_age = (
            False,
            False,
            False,
            False,
        )
        df = pd.DataFrame()
        for table in df_files:
            source_df = db.tables[table].df
            # check if columns should be renamed
            source_df = self._check_cols(source_df)
            # create a dataframe with the index (the filenames)
            df_local = pd.DataFrame(index=source_df.index)
            # try to get the targets from this dataframe
            try:
                # try to get the target values
                df_local[self.target] = source_df[self.col_label]
                is_labeled = True
            except (KeyError, ValueError, audformat.errors.BadKeyError):
                pass
            try:
                # try to get the speaker values
                df_local["speaker"] = source_df["speaker"]
                got_speaker = True
            except (KeyError, ValueError, audformat.errors.BadKeyError):
                pass
            try:
                # try to get the gender values
                if "gender" in source_df:
                    df_local["gender"] = source_df["gender"]
                    got_gender = True
            except (KeyError, ValueError, audformat.errors.BadKeyError):
                pass
            try:
                # try to get the age values
                df_local["age"] = source_df["age"].astype(int)
                got_age = True
            except (KeyError, ValueError, audformat.errors.BadKeyError):
                pass
            try:
                # also it might be possible that the sex is part of the speaker description
                df_local["gender"] = db[table]["speaker"].get(map="gender")
                got_gender = True
            except (ValueError, audformat.errors.BadKeyError):
                pass
            try:
                # also it might be possible that the sex is part of the speaker description
                df_local["gender"] = db[table]["speaker"].get(map="sex")
                got_gender = True
            except (ValueError, audformat.errors.BadKeyError):
                pass
            try:
                # also it might be possible that the age is part of the speaker description
                df_local["age"] = db[table]["speaker"].get(map="age").astype(int)
                got_age = True
            except (ValueError, audformat.errors.BadKeyError):
                pass
            try:
                # same for the target, e.g. "age"
                df_local[self.target] = db[table]["speaker"].get(map=self.target)
                is_labeled = True
            except (ValueError, audformat.core.errors.BadKeyError):
                pass
            # copy other column
            for column in source_df.columns:
                if column not in [self.target, "age", "speaker", "gender"]:
                    df_local[column] = source_df[column]
            df = pd.concat([df, df_local])
        return df, is_labeled, got_speaker, got_gender, got_age

    def split(self):
        """Split the datbase into train and development set"""
        store = self.util.get_path("store")
        storage_test = f"{store}{self.name}_testdf.pkl"
        storage_train = f"{store}{self.name}_traindf.pkl"
        split_strategy = self.util.config_val_data(
            self.name, "split_strategy", "speaker_split"
        )
        self.util.debug(
            f"splitting database {self.name} with strategy {split_strategy}"
        )
        # 'database' (default), 'speaker_split', 'specified', 'reuse'
        if split_strategy != "speaker_split" and not self.start_fresh:
            # check if the splits have been computed previously (not for speaker split)
            if os.path.isfile(storage_train) and os.path.isfile(storage_test):
                # if self.util.config_val_data(self.name, 'test_tables', False):
                self.util.debug(
                    "splits: reusing previously stored test file" f" {storage_test}"
                )
                self.df_test = pd.read_pickle(storage_test)
                self.util.debug(
                    "splits: reusing previously stored train file" f" {storage_train}"
                )
                self.df_train = pd.read_pickle(storage_train)

                return
            elif os.path.isfile(storage_train):
                self.util.debug(
                    "splits: reusing previously stored train file" f" {storage_train}"
                )
                self.df_train = pd.read_pickle(storage_train)
                self.df_test = pd.DataFrame()
                return
            elif os.path.isfile(storage_test):
                self.util.debug(
                    "splits: reusing previously stored test file" f" {storage_test}"
                )
                self.df_test = pd.read_pickle(storage_test)
                self.df_train = pd.DataFrame()
                return
        if split_strategy == "database":
            #  use the splits from the database
            testdf = self.db.tables[self.target + ".test"].df
            traindf = self.db.tables[self.target + ".train"].df
            # use only the train and test samples that were not perhaps filtered out by an earlier processing step
            self.df_test = self.df.loc[self.df.index.intersection(testdf.index)]
            self.df_train = self.df.loc[self.df.index.intersection(traindf.index)]
        elif split_strategy == "train":
            self.df_train = self.df
            self.df_test = pd.DataFrame()
        elif split_strategy == "test":
            self.df_test = self.df
            self.df_train = pd.DataFrame()
        elif split_strategy == "specified":
            traindf, testdf = pd.DataFrame(), pd.DataFrame()
            # try to load some dataframes for testing
            entry_test_tables = self.util.config_val_data(
                self.name, "test_tables", False
            )
            if entry_test_tables:
                test_tables = ast.literal_eval(entry_test_tables)
                for test_table in test_tables:
                    testdf = pd.concat([testdf, self.db.tables[test_table].df])
            entry_train_tables = self.util.config_val_data(
                self.name, "train_tables", False
            )
            if entry_train_tables:
                train_tables = ast.literal_eval(entry_train_tables)
                for train_table in train_tables:
                    traindf = pd.concat([traindf, self.db.tables[train_table].df])
            # use only the train and test samples that were not perhaps filtered out by an earlier processing step
            # testdf.index.map(lambda x: os.path.join(self.root, x))
            #            testdf.index = testdf.index.to_series().apply(lambda x: self.root+x)
            testdf = testdf.set_index(
                audformat.utils.to_segmented_index(testdf.index, allow_nat=False)
            )
            traindf = traindf.set_index(
                audformat.utils.to_segmented_index(traindf.index, allow_nat=False)
            )
            self.df_test = self.df.loc[self.df.index.intersection(testdf.index)]
            self.df_train = self.df.loc[self.df.index.intersection(traindf.index)]
            # it might be necessary to copy the target values
            try:
                self.df_test[self.target] = testdf[self.target]
            except KeyError:
                pass  # if the dataframe is empty
            try:
                self.df_train[self.target] = traindf[self.target]
            except KeyError:
                pass  # if the dataframe is empty
        elif split_strategy == "balanced":
            self.balanced_split()
        elif split_strategy == "speaker_split":
            self.split_speakers()
        elif split_strategy == "random":
            self.random_split()
        elif split_strategy == "reuse":
            self.util.debug(f"{self.name}: trying to reuse data splits")
            self.df_test = pd.read_pickle(storage_test)
            self.df_train = pd.read_pickle(storage_train)
        elif isinstance(ast.literal_eval(split_strategy), list):
            # treat this as a list of test speakers
            self.assign_speakers(ast.literal_eval(split_strategy))
        else:
            self.util.error(f"unknown split strategy: {split_strategy}")

        # check if train or test set should be ignored
        as_test = eval(self.util.config_val_data(self.name, "as_test", "False"))
        if as_test:
            self.df_train = pd.DataFrame()
        as_train = eval(self.util.config_val_data(self.name, "as_train", "False"))
        if as_train:
            self.df_test = pd.DataFrame()

        if self.df_test.shape[0] > 0:
            self.df_test = self.finish_up(self.df_test, storage_test)
        if self.df_train.shape[0] > 0:
            self.df_train = self.finish_up(self.df_train, storage_train)

        self.util.debug(
            f"{self.name}: {self.df_test.shape[0]} samples in test and"
            f" {self.df_train.shape[0]} samples in train"
        )

    def finish_up(self, df, storage):
        """
        Bin target values if they are continuous but a classification experiment should be done
        self.check_continuous_classification(df)
        remember the splits for future use
        """
        df.is_labeled = self.is_labeled
        self.df_test.is_labeled = self.is_labeled
        df.to_pickle(storage)
        return df

    def balanced_split(self):
        """One way to split train and eval sets: Generate split dataframes for some balancing criterion"""
        from splitutils import binning, optimize_traintest_split

        seed = 42
        k = 30
        test_size = int(self.util.config_val_data(self.name, "test_size", 20)) / 100.0
        df = self.df
        # split target
        targets = df[self.target].to_numpy()
        #
        bins = self.util.config_val("DATA", "bin", False)
        if bins:
            nbins = len(ast.literal_eval(bins))
            targets = binning(targets, nbins=nbins)
        # on which variable to split
        speakers = df["speaker"].to_numpy()

        # on which variables (targets, groupings) to stratify
        stratif_vars = self.util.config_val("DATA", "balance", False)
        stratif_vars_array = {}
        if not stratif_vars:
            self.util.error("balanced split needs stratif_vars to stratify the splits")
        else:
            stratif_vars = ast.literal_eval(stratif_vars)
            for stratif_var in stratif_vars.keys():
                if stratif_var == self.target:
                    stratif_vars_array[self.target] = targets
                    continue
                else:
                    data = df[stratif_var].to_numpy()
                    bins = self.util.config_val("DATA", f"{stratif_var}_bins", False)
                    if bins:
                        data = binning(data, nbins=int(bins))
                    stratif_vars_array[stratif_var] = data
        # weights for all stratify_on variables and
        # and for test proportion match. Give target
        # variable EMOTION more weight than groupings.
        size_diff = int(self.util.config_val("DATA", "size_diff_weight", "1"))
        weights = {
            "size_diff": size_diff,
        }
        for key, value in stratif_vars.items():
            weights[key] = value
        # find optimal test indices TEST_I in DF
        # info: dict with goodness of split information

        train_i, test_i, info = optimize_traintest_split(
            X=df,
            y=targets,
            split_on=speakers,
            stratify_on=stratif_vars_array,
            weight=weights,
            test_size=test_size,
            k=k,
            seed=seed,
        )
        self.util.debug(f"stratification info;\n{info}")
        self.df_train = df.iloc[train_i]
        self.df_test = df.iloc[test_i]
        self.util.debug(
            f"{self.name} (balanced split): [{self.df_train.shape[0]}/{self.df_test.shape[0]}]"
            " samples in train/test"
        )
        # because this generates new train/test sample quantaties, the feature extraction has to be done again
        glob_conf.config["FEATS"]["needs_feature_extraction"] = "True"

    def assign_speakers(self, speakers):
        """One way to split train and eval sets: Specify test speaker names."""
        self.df_test = self.df[self.df.speaker.isin(speakers)]
        if len(self.df_test) == 0:
            self.util.error(f"no speakers found in {speakers}")
        self.df_train = self.df[~self.df.index.isin(self.df_test.index)]
        self.util.debug(
            f"{self.name} (speakers assigned): [{self.df_train.shape[0]}/{self.df_test.shape[0]}]"
            " samples in train/test"
        )
        # because this generates new train/test sample quantaties, the feature extraction has to be done again
        glob_conf.config["FEATS"]["needs_feature_extraction"] = "True"

    def split_speakers(self):
        """One way to split train and eval sets: Specify percentage of evaluation speakers"""
        test_percent = int(self.util.config_val_data(self.name, "test_size", 20))
        df = self.df
        s_num = df.speaker.nunique()
        test_num = int(s_num * (test_percent / 100))
        test_spkrs = sample(list(df.speaker.unique()), test_num)
        self.df_test = df[df.speaker.isin(test_spkrs)]
        self.df_train = df[~df.index.isin(self.df_test.index)]
        self.util.debug(
            f"{self.name} (speaker split): [{self.df_train.shape[0]}/{self.df_test.shape[0]}]"
            " samples in train/test"
        )
        # because this generates new train/test sample quantaties, the feature extraction has to be done again
        try:
            glob_conf.config["FEATS"]["needs_feature_extraction"] = "True"
        except KeyError:
            pass

    def random_split(self):
        """One way to split train and eval sets: Specify percentage of random samples"""
        test_percent = int(self.util.config_val_data(self.name, "test_size", 20))
        df = self.df
        s_num = len(df)
        test_num = int(s_num * (test_percent / 100))
        test_smpls = sample(list(df.index), test_num)
        self.df_test = df[df.index.isin(test_smpls)]
        self.df_train = df[~df.index.isin(self.df_test.index)]
        self.util.debug(
            f"{self.name}: [{self.df_train.shape[0]}/{self.df_test.shape[0]}]"
            " samples in train/test"
        )
        # because this generates new train/test sample quantaties, the feature extraction has to be done again
        glob_conf.config["FEATS"]["needs_feature_extraction"] = "True"

    def _add_labels(self, df):
        df.is_labeled = self.is_labeled
        df.got_gender = self.got_gender
        df.got_age = self.got_age
        df.got_speaker = self.got_speaker
        return df

    def prepare_labels(self):
        # strategy = self.util.config_val("DATA", "strategy", "train_test")
        only_tests = eval(self.util.config_val("DATA", "tests", "False"))
        module = glob_conf.module
        if only_tests and module == "test":
            self.df_test = self.map_labels(self.df_test)
            # Bin target values if they are continuous but a classification experiment should be done
            self.map_continuous_classification(self.df_test)
            self.df_test = self._add_labels(self.df_test)
            if self.util.config_val_data(self.name, "value_counts", False):
                if not self.got_gender or not self.got_speaker:
                    self.util.error(
                        "can't plot value counts if no speaker or gender is" " given"
                    )
                else:
                    self.plot.describe_df(
                        self.name, self.df_test, self.target, f"{self.name}_distplot"
                    )
            return
        self.df_train = self.map_labels(self.df_train)
        self.df_test = self.map_labels(self.df_test)
        self.map_continuous_classification(self.df_train)
        self.map_continuous_classification(self.df_test)
        self.df_train = self._add_labels(self.df_train)
        self.df_test = self._add_labels(self.df_test)
        if self.util.config_val_data(self.name, "value_counts", False):
            if not self.got_gender or not self.got_speaker:
                self.util.error(
                    "can't plot value counts if no speaker or gender is" " given"
                )
            else:
                self.plot.describe_df(
                    self.name,
                    self.df_train,
                    self.target,
                    f"{self.name}_train_distplot",
                )
                self.plot.describe_df(
                    self.name,
                    self.df_test,
                    self.target,
                    f"{self.name}_test_distplot",
                )

    def map_labels(self, df):
        pd.options.mode.chained_assignment = None
        if (
            df.shape[0] == 0
            or not self.util.exp_is_classification()
            or self.check_continuous_classification()
        ):
            return df
        """Rename the labels and remove the ones that are not needed."""
        target = glob_conf.config["DATA"]["target"]
        # see if a special mapping should be used
        mappings = self.util.config_val_data(self.name, "mapping", False)
        if mappings:
            mapping = ast.literal_eval(mappings)
            df[target] = df[target].map(mapping)
            self.util.debug(f"{self.name}: mapped {mapping}")
        # remove labels that are not in the labels list
        labels = self.util.config_val("DATA", "labels", False)
        if labels:
            labels = ast.literal_eval(labels)
            df = df[df[target].isin(labels)]
        else:
            labels = list(df[target].unique())
        df["class_label"] = df[target]
        return df

    def check_continuous_classification(self):
        datatype = self.util.config_val("DATA", "type", False)
        if self.util.exp_is_classification() and datatype == "continuous":
            return True
        return False

    def map_continuous_classification(self, df):
        """Map labels to bins for continuous data that should be classified"""
        if self.check_continuous_classification():
            self.util.debug(f"{self.name}: binning continuous variable to categories")
            cat_vals = self.util.continuous_to_categorical(df[self.target])
            df[self.target] = cat_vals.values
            labels = ast.literal_eval(glob_conf.config["DATA"]["labels"])
            df["class_label"] = df[self.target]
            for i, l in enumerate(labels):
                df["class_label"] = df["class_label"].replace(i, str(l))



# transformer_feature_extractor.py

import os

import pandas as pd
import torch
import torchaudio
from tqdm import tqdm

import nkululeko.glob_conf as glob_conf
from nkululeko.feat_extract.featureset import Featureset


class TransformerFeatureExtractor(Featureset):
    def __init__(self, name, data_df, feat_type):
        super().__init__(name, data_df, feat_type)
        cuda = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = self.util.config_val("MODEL", "device", cuda)
        self.model_initialized = False
        self.feat_type = feat_type

    def init_model(self):
        raise NotImplementedError("Subclasses must implement init_model method")

    def get_embeddings(self, signal, sampling_rate, file):
        try:
            with torch.no_grad():
                # Preprocess the input
                inputs = self.preprocess_input(signal, sampling_rate)

                # Get model outputs
                outputs = self.model(**inputs)

                # Extract the relevant hidden states
                hidden_states = self.extract_hidden_states(outputs)

                # Pool the hidden states
                embeddings = self.pool_hidden_states(hidden_states)

                # Convert to numpy and flatten
                embeddings = embeddings.cpu().numpy().ravel()

            return embeddings

        except Exception as e:
            self.util.error(f"Error extracting embeddings for file {file}: {str(e)}")
            return np.zeros(self.get_embedding_dim())  # Return zero vector on error

    def preprocess_input(self, signal, sampling_rate):
        # This method should be implemented by subclasses
        raise NotImplementedError("Subclasses must implement preprocess_input method")

    def extract_hidden_states(self, outputs):
        # This method should be implemented by subclasses
        raise NotImplementedError(
            "Subclasses must implement extract_hidden_states method"
        )

    def pool_hidden_states(self, hidden_states):
        # Default implementation: mean pooling over time dimension
        return torch.mean(hidden_states, dim=1)

    def get_embedding_dim(self):
        # This method should be implemented by subclasses
        raise NotImplementedError("Subclasses must implement get_embedding_dim method")

    def extract(self):
        store = self.util.get_path("store")
        storage = f"{store}{self.name}.pkl"
        extract = self.util.config_val("FEATS", "needs_feature_extraction", False)
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if extract or no_reuse or not os.path.isfile(storage):
            if not self.model_initialized:
                self.init_model()
            self.util.debug(
                f"extracting {self.feat_type} embeddings, this might take a while..."
            )
            emb_series = pd.Series(index=self.data_df.index, dtype=object)
            for idx, (file, start, end) in enumerate(
                tqdm(self.data_df.index.to_list())
            ):
                signal, sampling_rate = torchaudio.load(
                    file,
                    frame_offset=int(start.total_seconds() * 16000),
                    num_frames=int((end - start).total_seconds() * 16000),
                )
                assert sampling_rate == 16000, f"got {sampling_rate} instead of 16000"
                emb = self.get_embeddings(signal, sampling_rate, file)
                emb_series[idx] = emb
            self.df = pd.DataFrame(emb_series.values.tolist(), index=self.data_df.index)
            self.df.to_pickle(storage)
            try:
                glob_conf.config["DATA"]["needs_feature_extraction"] = "false"
            except KeyError:
                pass
        else:
            self.util.debug(f"reusing extracted {self.feat_type} embeddings")
            self.df = pd.read_pickle(storage)
            if self.df.isnull().values.any():
                self.util.error(
                    f"got nan: {self.df.shape} {self.df.isnull().sum().sum()}"
                )

    def extract_sample(self, signal, sr):
        self.init_model()
        feats = self.get_embeddings(signal, sr, "no file")
        return feats

        # for each feature extractor
        # feats_ast.py

        # class Ast(TransformerFeatureExtractor):
        #     def preprocess_input(self, signal, sampling_rate):
        #         inputs = self.processor(signal.numpy(), sampling_rate=sampling_rate, return_tensors="pt")
        #         return {k: v.to(self.device) for k, v in inputs.items()}

        #     def extract_hidden_states(self, outputs):
        #         return outputs.last_hidden_state

        #     def get_embedding_dim(self):
        #         return self.model.config.hidden_size

        # # feats_wav2vec2.py

        # class Wav2vec2(TransformerFeatureExtractor):
        #     def preprocess_input(self, signal, sampling_rate):
        #         inputs = self.processor(signal, sampling_rate=sampling_rate, return_tensors="pt")
        #         return {k: v.to(self.device) for k, v in inputs.items()}

        #     def extract_hidden_states(self, outputs):
        #         return outputs.last_hidden_state

        #     def get_embedding_dim(self):
        #         return self.model.config.hidden_size

        # # feats_wavlm.py

        # class Wavlm(TransformerFeatureExtractor):
        #     def preprocess_input(self, signal, sampling_rate):
        #         inputs = self.processor(signal, sampling_rate=sampling_rate, return_tensors="pt")
        #         return {k: v.to(self.device) for k, v in inputs.items()}

        #     def extract_hidden_states(self, outputs):
        return outputs.last_hidden_state

    # def get_embedding_dim(self):
    #     return self.model.config.hidden_size



# main.py
# Demonstration code to use the ML-experiment framework

import sys

sys.path.append("./src")
import configparser

import experiment as exp
from util import Util


def main(config_file):
    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    util = Util()

    # create a new experiment
    expr = exp.Experiment(config)
    print(f"running {expr.name}")

    # load the experiment
    expr.load(f"{util.get_exp_name()}.pkl")

    expr.demo()

    print("DONE")


if __name__ == "__main__":
    main(sys.argv[1])



# scaler.py

import numpy as np
import pandas as pd
from sklearn.preprocessing import RobustScaler, StandardScaler

from nkululeko.utils.util import Util


class Scaler:
    """
    class to normalize speech features
    """

    def __init__(
        self, train_data_df, test_data_df, train_feats, test_feats, scaler_type
    ):
        """
        Initializer.

                Parameters:
                        train_data_df (pd.DataFrame): The training dataframe with speakers.
                            only needed for speaker normalization
                        test_data_df (pd.DataFrame): The test dataframe with speakers
                            only needed for speaker normalization
                        train_feats (pd.DataFrame): The train features dataframe
                        test_feats (pd.DataFrame): The test features dataframe (can be None)
        """
        self.util = Util("scaler")
        if scaler_type == "standard":
            self.scaler = StandardScaler()
        elif scaler_type == "robust":
            self.scaler = RobustScaler()
        elif scaler_type == "speaker":
            self.scaler = StandardScaler()
        elif scaler_type == "bins":
            pass
        else:
            self.util.error("unknown scaler: " + scaler_type)
        self.scaler_type = scaler_type
        self.feats_train = train_feats
        self.data_train = train_data_df
        self.feats_test = test_feats
        self.data_test = test_data_df

    def scale(self):
        """
        Actually scales/normalizes.

                Returns:
                        train_feats (pd.DataFrame): The scaled train features dataframe
                        test_feats (pd.DataFrame): The scaled test features dataframe (can be None)
        """
        if self.scaler_type != "speaker":
            self.util.debug("scaling features based on training set")
            return self.scale_all()
        else:
            self.util.debug("scaling features per speaker based on training")
            return self.speaker_scale()

    def scale_all(self):
        if self.scaler_type != "bins":
            self.scaler.fit(self.feats_train.values)
            self.feats_train = self.scale_df(self.feats_train)
            if self.feats_test is not None:
                self.feats_test = self.scale_df(self.feats_test)
        else:
            self.bin_to_three()
        return self.feats_train, self.feats_test

    def scale_df(self, df):
        scaled_features = self.scaler.fit_transform(df.values)
        df = pd.DataFrame(scaled_features, index=df.index, columns=df.columns)
        return df

    def speaker_scale(self):
        self.feats_train = self.speaker_scale_df(self.data_train, self.feats_train)
        if self.feats_test is not None:
            self.feats_test = self.speaker_scale_df(self.data_test, self.feats_test)
        return [self.feats_train, self.feats_test]

    def speaker_scale_df(self, df, feats_df):
        for speaker in df["speaker"].unique():
            indices = df.loc[df["speaker"] == speaker].index
            feats_df.loc[indices, :] = self.scaler.fit_transform(
                feats_df.loc[indices, :]
            )
        return feats_df

    def bin_to_three(self):
        feats_bin_train = pd.DataFrame(index=self.feats_train.index)
        feats_bin_test = pd.DataFrame(index=self.feats_test.index)
        for c in self.feats_train.columns:
            b1 = np.quantile(self.feats_train[c], 0.33)
            b2 = np.quantile(self.feats_train[c], 0.66)
            feats_bin_train[c] = self._bin(self.feats_train[c].values, b1, b2).values
            feats_bin_test[c] = self._bin(self.feats_test[c].values, b1, b2).values
        self.feats_train = feats_bin_train
        self.feats_test = feats_bin_test

    def _bin(self, series, b1, b2):
        bins = [-1000000, b1, b2, 1000000]
        labels = [0, 0.5, 1]
        result = np.digitize(series, bins) - 1
        result = pd.Series(result)
        for i, l in enumerate(labels):
            result = result.replace(i, str(l))
        return result



# mld_fset.py
import os
import sys

import nkululeko.glob_conf as glob_conf
from nkululeko.feat_extract.featureset import Featureset


class MLD_set(Featureset):
    def __init__(self, name, data_df):
        super().__init__(name, data_df)
        mld_path = self.util.config_val("FEATS", "mld.model", None)
        sys.path.append(mld_path)

    def extract(self):
        store = self.util.get_path("store")
        storage = f"{store}{self.name}.pkl"
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if no_reuse:
            os.remove(storage)
        if not os.path.isfile(storage):
            self.util.debug(
                "extracting midleveldescriptor features, this might take a" " while..."
            )
        else:
            self.util.debug("reusing previously extracted midleveldescriptor features")
        import midlevel_descriptors as mld

        fex_mld = mld.MLD()
        self.df = fex_mld.extract_from_index(index=self.data_df, cache_path=storage)
        self.util.debug(f"MLD feats shape: {self.df.shape}")
        # shouldn't happen
        # replace NANa with column means values
        self.util.debug("MLD extractor: checking for NANs...")
        for i, col in enumerate(self.df.columns):
            if self.df[col].isnull().values.any():
                self.util.debug(
                    f"{col} includes {self.df[col].isnull().sum()} nan,"
                    " inserting mean values"
                )
                self.df[col] = self.df[col].fillna(self.df[col].mean())

        try:
            # use only samples that have a minimum number of syllables
            min_syls = int(glob_conf.config["FEATS"]["min_syls"])
            self.df = self.df[self.df["hld_nSyl"] >= min_syls]
        except KeyError:
            pass
        if self.df.isna().to_numpy().any():
            self.util.error("feats 0: NANs exist")
        self.df = self.df.astype(float)
